# cython: language_level=3
from cpython.exc cimport PyErr_CheckSignals


cpdef object progressBar(object pos,
                  off_t maxi,
                  bint reset=False,
                  bytes head=b'',
                  list delta=[],
                  list step=[1,0,0]):
                  
    cdef off_t    ipos
    cdef double percent 
    cdef int days,hour,minu,sec
    cdef bytes bar
    cdef off_t fraction
    cdef int freq,cycle,arrow
    cdef tm remain

    cdef clock_t d
    cdef clock_t elapsed
    cdef clock_t newtime 
    cdef clock_t more
    
    #                   0123456789
    cdef char* wheel=  '|/-\\'
    cdef char*  spaces='          ' \
                       '          ' \
                       '          ' \
                       '          ' \
                       '          '
                
    cdef char*  diese ='##########' \
                       '##########' \
                       '##########' \
                       '##########' \
                       '##########' 
                  
    if reset:
        del delta[:]
        step[0]=1
        step[1]=0
        step[2]=0
    if not delta:
        delta.append(clock())
        delta.append(clock())
        
    if maxi<=0:
        maxi=1
    assert maxi>0
    
    freq,cycle,arrow = step

    cycle+=1
    
    if cycle % freq == 0:
        cycle=1
        newtime = clock()
        try:
            d = newtime-delta[1]
        except OverflowError:
            d=0
            
        if d < 0.02 * CLOCKS_PER_SEC :
            freq*=2
        elif d > 0.4 * CLOCKS_PER_SEC and freq>1:
            freq/=2
            
        delta[1]=newtime
        try:
            elapsed = newtime-delta[0]
        except OverflowError:
            elapsed = 0
        
        if callable(pos):
            ipos=pos()
        else:
            ipos=pos
            
        percent = <double>ipos/<double>maxi
        more = <time_t>((<double>elapsed / percent * (1. - percent))/CLOCKS_PER_SEC)
        <void>gmtime_r(&more, &remain)
        days = remain.tm_yday 
        hour = remain.tm_hour
        minu  = remain.tm_min
        sec  = remain.tm_sec

        fraction=<int>(percent * 50.)
        arrow=(arrow+1) % 4
        diese[fraction]=0
        spaces[50 - fraction]=0
        
        if days:
            <void>fprintf(stderr,b'\r%s %5.1f %% |%s%c%s] remain : %d days %02d:%02d:%02d',
                            <char*>head,
                            percent*100,
                            diese,wheel[arrow],spaces,
                            days,hour,minu,sec)
        else:
            <void>fprintf(stderr,b'\r%s %5.1f %% |%s%c%s] remain : %02d:%02d:%02d',
                            <char*>head,
                            percent*100.,
                            diese,wheel[arrow],spaces,
                            hour,minu,sec)
            
        diese[fraction]=b'#'
        spaces[50 - fraction]=b' '
        
        # Added to check if signal (Ctrl-C) are send to the program
        PyErr_CheckSignals()

    else:
        cycle+=1

    step[0:3] = freq,cycle,arrow



# cython: language_level=3

from ._graphmultiedge cimport *
from cpython.list cimport PyList_Size
from builtins import isinstance

cdef class DiGraphMultiEdge(DiGraph):

    cpdef dict addEdge(self,int node1, int node2):
        cdef char[50] buffer1
        cdef char[50] buffer2
        cdef bytes key1 
        cdef bytes key2 
        cdef tuple edge
        cdef list eattr
        cdef int eid
        
        snprintf(buffer1,50,b"%d",node1)
        key1 = buffer1
        
        snprintf(buffer2,50,b"%d",node2)
        key2 = buffer2
        
        edge = (key1,key2)
        
        if edge not in self._edges_attrs:
            self.addNode(node1)
            self.addNode(node2)
        
            self._nodes[key1][0].add(key2)
            self._nodes[key2][1].add(key1)

            self._edges_attrs[edge]=[]
            
        eattr = self._edges_attrs[edge]
        eattr.append({})
        #eid = PyList_Size(eattr)-1
            
        self._edgecount+=1
        
        return eattr[-1]
            
    cpdef dict getEdgeAttr(self, int node1, int node2,int edge=-1):
 
        cdef char[50] buffer1
        cdef char[50] buffer2
        cdef bytes key1 
        cdef bytes key2 
        cdef tuple edgek
        cdef list attr
        cdef dict rep
        
        snprintf(buffer1,50,b"%d",node1)
        key1 = buffer1
        
        snprintf(buffer2,50,b"%d",node2)
        key2 = buffer2
   
        edgek = (key1,key2)
        
        attr = self._edges_attrs[edgek]
        
        return attr[edge]
    
    cpdef int edgeCount(self):
        return self._edgecount
    
    cpdef deleteEdge(self, int node1, int node2, int edge=-1):
        cdef char[50] buffer1
        cdef char[50] buffer2
        cdef bytes key1 
        cdef bytes key2 
        cdef tuple edgek
        cdef list edgel

        
        snprintf(buffer1,50,b"%d",node1)
        key1 = buffer1
        
        snprintf(buffer2,50,b"%d",node2)
        key2 = buffer2
   
        edgek = (key1,key2)
        
        edgel = self._edges_attrs[edgek]
        
        del edgel[edge]
        self._edgecount-=1
        
        if not edgel:
            del self._edges_attrs[edgek]
            
            self._nodes[key1][0].remove(key2)
            self._nodes[key2][1].remove(key1)

    cpdef deleteAllEdges(self, int node1, int node2):
        cdef char[50] buffer1
        cdef char[50] buffer2
        cdef bytes key1 
        cdef bytes key2 
        cdef tuple edgek
        cdef list edgel
        cdef int nedge

        
        snprintf(buffer1,50,b"%d",node1)
        key1 = buffer1
        
        snprintf(buffer2,50,b"%d",node2)
        key2 = buffer2
   
        edgek = (key1,key2)
        nedge = len(self._edges_attrs[edgek])
        
        del self._edges_attrs[edgek]
        self._nodes[key1][0].remove(key2)
        self._nodes[key2][1].remove(key1)
       
        self._edgecount-=nedge
        
    cpdef deleteNode(self, int node):
        cdef char[50] buffer
        cdef bytes key
        cdef dict attr
        cdef int node2
        cdef list edges
        
        snprintf(buffer,50,b"%d",node)
        key = buffer
        
        del self._nodes_attrs[key]
        
        edges = list(self.neighbourIterator(node))
        for node2 in edges:
            self.deleteAllEdges(node,node2)
            
        edges = list(self.parentIterator(node))
        for node2 in edges:
            self.deleteAllEdges(node2,node)
            
        del self._nodes[key]


    def edgeIterator(self,nodePredicate=None, edgePredicate=None):
        cdef tuple edge
        cdef bytes key1
        cdef bytes key2
        cdef int   node1
        cdef int   node2
        cdef int   edgei
        cdef int  edgel
        
        for edge in self._edges_attrs:
            key1 = edge[0]
            key2 = edge[1]
            
            node1 = atol(key1)
            node2 = atol(key2)
            
            if nodePredicate is None or (nodePredicate(node1) and nodePredicate(node2)):
                edgel = len(self._edges_attrs[(key1,key2)])
                for edgei in range(edgel):
                    edge=(node1,node2,edgei)
                    if edgePredicate is None or edgePredicate(edge):
                        yield edge
                    
                
    def parentIterator(self,int node,nodePredicate=None, edgePredicate=None):
        cdef char[50] buffer
        cdef bytes key1
        cdef bytes key2
        cdef set neighbours
        cdef int n 
        cdef int edgel
        cdef int edgei
        cdef bint edgeok
        
        snprintf(buffer,50,b"%d",node)
        key1 = buffer
        
        neighbours = self._nodes[key1][1]
        
        for key2 in neighbours:
            n = atol(key2)
            if nodePredicate is None or nodePredicate(n):
                edgel = len(self._edges_attrs[(key2,key1)])
                edgeok = True
                if edgePredicate is not None :
                    for edgei in range(edgel):
                        edgeok=edgeok and edgePredicate((n,node,edgei))
                if edgePredicate is None or edgeok:
                    yield n
                
    def neighbourIterator(self,int node,nodePredicate=None, edgePredicate=None):
        cdef char[50] buffer
        cdef bytes key1
        cdef bytes key2
        cdef set neighbours
        cdef int n 
        cdef int edgel
        cdef int edgei
        cdef bint edgeok
        
        snprintf(buffer,50,b"%d",node)
        key1 = buffer
        
        neighbours = self._nodes[key1][0]
        
        for key2 in neighbours:
            n = atol(key2)
            if nodePredicate is None or nodePredicate(n):
                edgel = len(self._edges_attrs[(key1,key2)])
                edgeok = True
                if edgePredicate is not None :
                    for edgei in range(edgel):
                        edgeok=edgeok and edgePredicate((node,n,edgei))
                if edgePredicate is None or edgeok:
                    yield n
                    




import copy
from ._ahocorasick import AhoCorasick

UniversalGeneticCode = {"F" :("TTT","TTC"),
                        "L" :("TTA","TTG","TAG","CTA",'CTG',"CTC","CTT"),
                        "S" :("TCA","TCC","TCG","TCT","AGA","AGC","AGG","AGT"),
                        "Y" :("TAA","TAC","TAT"),
                        "C" :("TGT","TGC"),
                        "W" :("TGA","TGG"),
                        "P" :("CCA","CCG","CCT","CCC"),
                        "H" :("CAT","CAC"),
                        "Q" :("CAA","CAG"),
                        "R" :("CGA","CGC","CGG","CGT","AGA","AGG"),
                        "I" :("ATT","ATC","ATA"),
                        "M" :("ATA","ATG"),
                        "T" :("ACA","ACT","ACG","ACC"),
                        "N" :("AAT","AAC","AAA"),
                        "K" :("AAA","AAG"),
                        "V" :("GTA","GTC","GTG","GTT"),
                        "A" :("GCA","GCC","GCG","GCT"),
                        "D" :("GAT","GAC"),
                        "E" :("GAA","GAG"),
                        "G" :("GGA","GGC","GGG","GGT")}

def listCodons2Dict(codons,offset=0,protein='prot'):
    d = [[(protein,offset)],None,{}]

    for c in codons:
        s = d[2]
        p = 0
        for l in c:
            s[l] = s.get(l,[[(protein,p+offset)],None,{}])
            s=s[l][2]
            p+=1
    
    return d[2]
        
def connect2codons(codon1,codon2,prot='prot'): 
    
    for l1 in codon1:
        for l2 in codon1[l1][2]:
            for l3 in codon1[l1][2][l2][2]:
                codon1[l1][2][l2][2][l3][2]=copy.deepcopy(codon2)
                
def mergeAutomata(automata1,automata2):
    
    for k in automata2:
        if k in automata1:
            automata1[k][0].extend(automata2[k][0])
            mergeAutomata(automata1[k][2], automata2[k][2])
        else:
            automata1[k]=copy.deepcopy(automata2[k])

def cleanAutomata(automata,length=100000,keepOnlyTerminal=True):
    lword=0
    d = automata
    
    stack=[(lword,d)]
    
    while stack:
        lword,d = stack.pop(0)
        lword+=1
        for val in d.itervalues():
            if lword==length:
                val[2]={}
            else:
                stack.append((lword,val[2]))
                
            # Keep only leaves as terminal state 
                
            if keepOnlyTerminal and val[2]:
                val[0]=[]

def buildAutomata(sequence,prot="prot",kup=4):
    
    automata = {}
    kup+=1
        
    for pos in xrange(len(sequence)-kup+1):
        peptide = sequence[pos:(pos+kup)]
        print(peptide)
        bt= [listCodons2Dict(UniversalGeneticCode[peptide[x]],
                             (pos+x)*3,
                             prot) 
             for x in xrange(kup)
            ]
        for shift in xrange(kup-1,0,-1):
            print(shift)
            connect2codons(bt[shift-1],bt[shift],prot) 
            
        autopep = bt[0]
        mergeAutomata(automata, autopep)
        for k in autopep:
            autopep2 = autopep[k][2]
            mergeAutomata(automata, autopep2)
            for k2 in autopep2:
                mergeAutomata(automata, autopep2[k2][2])
                
    cleanAutomata(automata, (kup-1)*3)
                    
    return automata
            
        
        
    

def automataIterator(automata):
    word=""
    d = automata[1]
    
    stack=[(word,d)]
    
    while stack:
        word,d = stack.pop(0)
        for k,val in d.iteritems():
            if k!='*':
                seq = word+k
            else:
                seq = word
            if isinstance(val, dict):
                stack.append((seq,val))
            if isinstance(val, list):
                yield seq,val

                       
def setLastCodon(codon1,prot='prot'):
    for l1 in codon1[1]:
        if l1!='*':
            for l2 in codon1[1][l1]:
                if l2!='*':
                    for l3 in codon1[1][l1][l2]:
                        if l3!='*':
                            codon1[1][l1][l2][l3]=[(prot,codon1[0]*3+2),None]
                        else:
                            codon1[1][l1][l2][l3]=[(prot,codon1[0]*3+1),None]
                else:
                    codon1[1][l1][l2]=[(prot,codon1[0]*3),None]
        else:
            codon1[1][l1]=[(prot,codon1[0]*3-1),None]

#def buildAutomata(sequence,prot='prot'):
#    bt = [listCodons2Dict(UniversalGeneticCode[sequence[x]],x*3,prot) for x in xrange(len(sequence))]
#    
#    setLastCodon(bt[-1], prot)
#    
#    for x in xrange(len(sequence)-1):
#        connect2codons(bt[-x-2], bt[-x-1], prot)
#        
#    return bt[0]

                
def matchSuffix(automata,suffixe):
    d = automata[1]
    
    for i in xrange(len(suffixe)):
        l = suffixe[i]
        if l in d:
            d=d[l]
        else:
            return None
    return d 

def longestSuffix(automata,seq):
    if seq:
        seq=seq[1:]
        
    d = matchSuffix(automata,seq)
    
    while not isinstance(d, dict):
        seq=seq[1:]
        d = matchSuffix(automata,seq)
        
    return seq,d

def buildErrorLink(automata):
    word=""
    d = automata[1]
    
    stack=[(word,d)]
    
    while stack:
        word,d = stack.pop(0)
        print >>sys.stderr,"Seq length : %d \r" % len(word),
        for k,val in d.iteritems():
            if k!='*':
                seq = word+k
            else:
                seq = word
            if isinstance(val, dict):
                stack.append((seq,val))
            if isinstance(val, list):
                val[1]=longestSuffix(automata, word)[1]
                



'''
Created on 24 mai 2015

@author: coissac
'''

from distutils import util

def is_mac_system():
    platform = util.get_platform().split('-')[0]
    
    return platform=='macosx'

def is_windows_system():
    platform = util.get_platform().split('-')[0]
    
    return platform=='Windows'
        


'''
Created on 28 sept. 2014

@author: coissac
'''
import sys
from orgasm import getOutput,getIndex, getSeeds
from orgasm.tango import restoreGraph, estimateFragmentLength, genesincontig,\
    pathConstraints, scaffold, selectGoodComponent, unfoldAssembling, path2fasta, \
    parseFocedScaffold
    
from orgasm.format.sam import path2sam,buildReadSequence
from orgasm.version import version

__title__="Universal assembling graph unfolder"

default_config = { 'circular' : False,
                   'force'    : False,
                   'format'   : "fasta",
                   "fscaffold": [],
                   "path"     : None,
                   "contigs"  : False,
                   'tags'      : None,
                   'extension' : None,
                   '1x-coverage' : None
                 }

def addOptions(parser):
    parser.add_argument(dest='orgasm:indexfilename',  metavar='index', 
                        help='index root filename (produced by the oa index command)')
    
    parser.add_argument(dest='orgasm:outputfilename',     metavar='output', 
                                                          nargs='?', 
                                                          default=None,
                        help='output prefix' )
    
    
    parser.add_argument('--path',             dest='unfold:path', 
                                              action='store',
                                              metavar='<edgeid>',
                                              type=int, 
                                              nargs='+', 
                                              default=None, 
                        help='A list of edge id separated by space add -- at the end of the path')

    parser.add_argument('--contigs',         dest='unfold:contigs', 
                                              action='store_true', 
                                              default=None, 
                        help='Outputs each stem of the assembly graph')

    parser.add_argument('--circular',         dest='unfold:circular', 
                                              action='store_true', 
                                              default=None, 
                        help='Wish a circular unfolding')

    parser.add_argument('--sam',         dest='unfold:format', 
                                              action='store_const', 
                                              const="sam",
                                              default=None, 
                        help='Produce a SAM file of the assembly')

    parser.add_argument('--fasta',            dest='unfold:format', 
                                              action='store_const', 
                                              const="fasta",
                                              default=None, 
                        help='Produce a FASTA file of the assembly [default format]')

    parser.add_argument('--force',            dest='unfold:force', 
                                              action='store_true', 
                                              default=None, 
                        help='Force circular unfolding')

    parser.add_argument('--force-scaffold',   dest='unfold:fscaffold', 
                                              action='append', 
                                              default=None, 
                        help='Force circular unfolding')

    parser.add_argument('--back',             dest='orgasm:back', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='the number of bases taken at the end of '
                             'contigs to jump with paired-ends [default: <estimated>]')
    
    parser.add_argument('--1x-coverage',      dest='unfold:1x-coverage', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='The expected coverage for not duplicated regions [default: <estimated>]')

    parser.add_argument('--set-tag','-S',     dest ='unfold:tags', 
                                              metavar='tag', 
                                              action='append',
                                              default=[], 
                                              type=str, 
                        help='Allows to add a tag in the OBITools format '
                             'to the header of the fasta sequences')



def buildPath(asm,compactgraph,back,minlink,config):
    logger=config['orgasm']['logger']
    logger.info("Select the good connected components")
    gcc = selectGoodComponent(compactgraph)
    
    if config['unfold']['circular']:
        if config['unfold']['force']:
            logger.info("Force circular sequence")
        else:
            logger.info("Unfolding in circular mode")

    constraints = pathConstraints(asm,compactgraph,
                                  back=int(back),minlink=minlink)

    paths = []
    for seeds in gcc:
        path = unfoldAssembling(asm,compactgraph,
                                seeds=seeds,
                                constraints=constraints,
                                circular=config['unfold']['circular'],
                                force=config['unfold']['force'],
                                cov1x=config['unfold']['1x-coverage'],
                                logger=logger)
            
        paths.append(path[-1][0])
        
    return(paths)
        
def contigPath(asm,compactgraph,config):
    edges = [compactgraph.getEdgeAttr(*i)['stemid']
              for i in compactgraph.edgeIterator(edgePredicate = lambda e : compactgraph.getEdgeAttr(*e)['class']=="sequence")]
    return [(i,) for i in edges if i > 0]
   
    
def run(config):

    logger=config['orgasm']['logger']
    output = getOutput(config)

    if config['unfold']['force']:
        config['unfold']['circular']= True
        
    forcedscaffold = parseFocedScaffold(config['unfold']['fscaffold'])
    
         
    r = getIndex(config)
    coverage,x,newprobes = getSeeds(r,config)  # @UnusedVariable
    
    asm = restoreGraph(output+'.oax',r,x)

    logger.info("Evaluate fragment length")
    
    meanlength,sdlength = estimateFragmentLength(asm)
    
    if meanlength is not None:
        logger.info("Fragment length estimated : %f pb (sd: %f)" % (meanlength,sdlength))

    if config['orgasm']['back'] is not None:
        back = config['orgasm']['back']
    elif config['orgasm']['back'] is None and meanlength is not None:
        back = int(meanlength + 4 * sdlength)
        if back > 500:
            back=500
    else:
        back = 300
        
    logger.info("Evaluate pair-end constraints")
    
    cg = asm.compactAssembling(verbose=False)
    
    genesincontig(cg,r,x)

    minlink=config['orgasm']['minlink']
    scaffold(asm,cg,minlink=minlink,
             back=int(back),
             addConnectedLink=False,
             forcedLink=forcedscaffold,
             logger=logger)
     

    #fastaout = open(output+".fasta","w")
    fastaout = sys.stdout
    pathout  = open(output+".path","w")
    
    paths=[]
    #### --->
    if config['unfold']['path'] is not None:
        paths.append(tuple(config['unfold']['path']))
        
    if config['unfold']['contigs']:
        paths.extend(contigPath(asm,cg,config))
    
    if not paths:
        paths=buildPath(asm, cg, back, minlink, config)
    
    logger.info("Print the result as a %s file" % config['unfold']['format'])
    
    paths.sort()
       
    seqid = config['orgasm']['outputfilename'].split('/')[-1]

    if config['unfold']['format']=="sam":
        c=1
        printedPath=set()

        print("@HD\tVN:1.5\tSO:coordinate",file=fastaout)
        print("@PG\tID:ORG.Asm\tVN:%s" % version)
        
        for path in paths:
            rpath = tuple(-i for i in reversed(path))
            if path not in printedPath and rpath not in printedPath:
                printedPath.add(tuple(path))
                identifier="%s_%d" % (seqid,c)
                seqlength = len(buildReadSequence(asm,cg,identifier=identifier,
                                                  path=path,back=back,logger=logger))
                                
                print("@SQ\tSN:%s\tLN:%d" % (identifier,seqlength)) # <=== !!! ATTENTION !!!
                c+=1

    c=1
    printedPath=set()
    for path in paths:
       
        rpath = tuple(-i for i in reversed(path))
        
        if path not in printedPath and rpath not in printedPath:
            
            printedPath.add(tuple(path))
        
            logger.info("Expanded path : %s" % str(path))

            if config['unfold']['format']=="fasta":                   
                fa = path2fasta(asm,cg,path,
                     identifier="%s_%d" % (seqid,c),
                     back=back,
                     minlink=config['orgasm']['minlink'],
                     logger=logger,
                     tags=config['unfold']['tags'])
            if config['unfold']['format']=="sam":                   
                fa = path2sam(asm,cg,path,
                     identifier="%s_%d" % (seqid,c),
                     back=back,
                     minlink=config['orgasm']['minlink'],
                     logger=logger,
                     tags=config['unfold']['tags'])
    
            print(fa,file=fastaout)
            print(" ".join([str(x) for x in path]),file=pathout)

            c+=1
        
    with open(output +'.path.gml','w') as gmlfile:
        print(cg.gml(),file=gmlfile)

              



#cython: language_level=3

'''
Created on 27 mars 2016

@author: coissac
'''

import pkgutil

from .. import commands

cdef object loadCommand(str name,loader):
    '''
    Load a command module from its name and an ImpLoader
    
    This function is for internal use
    
    @param name:   name of the module
    @type name: str 
    @param loader: the module loader
    @type loader: ImpLoader
    
    @return the loaded module
    @rtype: module 
    '''
    
    module = loader.find_module(name).load_module(name)
    return module

def getCommandsList():
    '''
    Returns the list of sub-commands available to the main `obi` command
    
    @return: a dict instance with key corresponding to each command and
             value corresponding to the module
             
    @rtype: dict
    '''
    
    cdef dict cmds = dict((x[1],loadCommand(x[1],x[0])) 
                           for x in pkgutil.iter_modules(commands.__path__) 
                           if not x[2])
    return cmds



'''
Created on 6 oct. 2014

@author: coissac
'''

# try:
#     from setuptools.command.install import install as install_ori
# except ImportError:
#     from distutils.command.install import install as install_ori

from distutils.command.install import install as install_ori

class install(install_ori):
    
    def __init__(self,dist):
        install_ori.__init__(self, dist)
#        self.sub_commands.insert(0, ('build',lambda self: True))
        self.sub_commands.append(('install_sphinx',lambda self: self.distribution.serenity))



# cython: language_level=3
from .fakereads cimport FakeReads
from cpython cimport array as c_array

cdef extern from *:
    ctypedef char* const_char_p "const char*"

cdef extern from "inttypes.h":
    ctypedef int int32_t
    ctypedef unsigned int uint32_t


cdef extern from "string.h":
    cdef int  strncmp(const_char_p s1, const_char_p s2, size_t n)
    
cdef extern from "orgasm.h":
    ctypedef struct buffer_t:
        size_t         readSize     # Size of one read in base pair
        size_t         recordSize   # Size in bytes of one compressed read
        char*          records      # the start of the first record [0].
        size_t         readCount    # count of reads. One pair of reads counts for 2
        uint32_t*      order1       # a pointer used to point on a int array indicating an order
                                    # over the records. @IndentOk
        uint32_t*      order2       # a pointer used to point on a int array indicating an order
                                    # over the records.
        uint32_t*      complement

    ctypedef unsigned char* pnuc

cdef class Index:

    cdef buffer_t* _index
    cdef FakeReads _fakes

    cpdef bint contains(self, bytes word)
    cpdef int count(self, bytes word)
    cpdef int len(self)
    cpdef bytes getRead(self, int32_t readid, int32_t begin, int32_t length)
    cpdef dict getExtensions(self, bytes word)
    cpdef int getReadSize(self)
    cpdef tuple getReadIds(self, bytes seq)
    cpdef bint isFake(self,int32_t id)
    cpdef tuple getIds(self, int32_t id)
    cpdef int32_t getPairedRead(self, int32_t id)
    cpdef tuple normalizedPairedEndsReads(self, int32_t id)
    cpdef dict lookForSeeds(self, dict sequences, 
                            int kup=?, int mincov=?,
                            float identity=?,
                            object logger=?)
    cdef c_array.array lookForString(self, bytes word)

    cpdef dict checkedExtensions(self, bytes  probe,
                                       tuple  adapters5=?,
                                       tuple  adapters3=?,
                                       int    minread=?,
                                       double minratio=?,
                                       int    mincov=?,
                                       int    minoverlap=?,
                                       int    extlength=?,
                                       bint   lowfilter=?,
                                       object restrict=?,
                                       bint   exact=?)



'''
Created on 28 sept. 2014

@author: coissac
'''

import orgasm.samples

from orgasm import getOutput,getIndex, getSeeds, getAdapters
from orgasm.tango import matchtoseed, cutLowCoverage, cutSNPs,\
    estimateDeadBrancheLength, coverageEstimate, estimateFragmentLength,\
    genesincontig, scaffold, fillGaps, dumpGraph, restoreGraph

from orgasm.assembler import Assembler,tango,resetusedreads
import sys
import os.path


__title__="Build the initial assembling graph"


default_config = {   'reformat'      : False,
                     'minread'       : None,
                     'coverage'      : None,
                     'minratio'      : None,
                     'mincov'        : 1,
                     'minoverlap'    : 50,
                     'smallbranches' : None,
                     'lowcomplexity' : False,
                     'snp'           : False,
                     'assmax'        : 500000,      # Maximum size of the assembling
                     'testrun'       : 15000,       # length of to assembling to estimate coverage
                     'clean'         : False,
                     'forceseeds'    : False,
                     'maxfillgaps'   : -1,
                     'covratio'      : 3,
                     'covratiofill'  : 12
                 }

def addOptions(parser):
    parser.add_argument(dest='orgasm:indexfilename',  metavar='index', 
                        help='index root filename (produced by the oa index command)')
    
    parser.add_argument(dest='orgasm:outputfilename',     metavar='output', 
                                                          nargs='?', 
                                                          default=None,
                        help='output prefix' )
    
    
    
    parser.add_argument("--reformat",
                        dest="buildgraph:reformat",
                        action='store_true',
                        default=None,
                        help='Asks for reformatting an old assembling to the new format'
                       )

    parser.add_argument('--minread',          dest='buildgraph:minread', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='the minimum count of read to consider [default: <estimated>]')
    
    parser.add_argument('--coverage',         dest='buildgraph:coverage', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='the expected sequencing coverage [default: <estimated>]')
    
    parser.add_argument('--coverage-ratio',   dest='buildgraph:covratio', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='the ratio between the expected and the minimum coverage [default 3]')
    
    parser.add_argument('--fillgaps-ratio',   dest='buildgraph:covratio', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='the ratio between the expected and the minimum coverage during gap filling [default 12]')
    
    parser.add_argument('--minratio',         dest='buildgraph:minratio', 
                                              type=float, action='store', 
                                              default=None, 
                        help='minimum ratio between occurrences of an extension'
                             ' and the occurrences of the most frequent extension '
                             'to keep it. [default: <estimated>]')
    
    parser.add_argument('--mincov',           dest='buildgraph:mincov', 
                                              type=int, 
                                              action='store', 
                                              default=1, 
                        help='minimum occurrences of an extension to '
                             'keep it. [default: %(default)d]')
    
    parser.add_argument('--assmax',           dest='buildgraph:assmax', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='maximum base pair assembled')
    
    parser.add_argument('--maxfillgaps',      dest='buildgraph:maxfillgaps', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='maximum cycles of fillgaps')
    
    parser.add_argument('--minoverlap',       dest='buildgraph:minoverlap', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='minimum length of the overlap between '
                             'the sequence and reads to participate in '
                             'the extension. [default: <estimated>]')
    
    parser.add_argument('--smallbranches',    dest='buildgraph:smallbranches', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='maximum length of the branches to cut during '
                             'the cleaning process [default: <estimated>]')
    
    parser.add_argument('--lowcomplexity',    dest='buildgraph:lowcomplexity', 
                                              action='store_true', 
                                              default=False, 
                        help='Use also low complexity probes')
    
    parser.add_argument('--back',             dest='orgasm:back', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='the number of bases taken at the end of '
                             'contigs to jump with pared-ends [default: <estimated>]')
    
    parser.add_argument('--snp',              dest='buildgraph:snp', 
                                              action='store_true', 
                                              default=False, 
                        help='activate the SNP clearing mode')
    
    parser.add_argument('--adapt5',           dest ='orgasm:adapt5', 
                                              metavar='adapt5', 
                                              default='adapt5ILLUMINA', 
                                              type=str, 
                                              required=False,
                        help='adapter sequences used to filter reads beginning by such sequences'
                             '; either a fasta file containing '
                             'adapter sequences or internal set of adapter sequences '
                             'among %s' % (str(list(filter(lambda s: s.startswith('adapt5'),dir(orgasm.samples)))),) +' [default: %(default)s]' )

    parser.add_argument('--adapt3',           dest ='orgasm:adapt3', 
                                              metavar='adapt3', 
                                              default='adapt3ILLUMINA', 
                                              type=str, 
                                              required=False,
                        help='adapter sequences used to filter reads ending by such sequences'
                             '; either a fasta file containing '
                             'adapter sequences or internal set of adapter sequences '
                             'among %s' % (str(list(filter(lambda s: s.startswith('adapt3'),dir(orgasm.samples)))),) +' [default: %(default)s]' )

    parser.add_argument('--probes',            dest ='orgasm:seeds', 
                                              metavar='seeds', 
                                              action='append',
                                              default=[], 
                                              type=str, 
                        help='protein or nucleic seeds; either a fasta file containing '
                        'seed sequences or the name of one of the internal set of seeds '
                        'among %s' % str(list(filter(lambda s: s.startswith('prot') or 
                                                s.startswith('nuc'),dir(orgasm.samples)))))

    parser.add_argument('--phiX',              dest='orgasm:phix', 
                                              action='store_true', 
                                              default=None, 
                        help='activate the filtering of Phi-X174 sequences [default]')
    
    parser.add_argument('--phiX-off',              dest='orgasm:phix', 
                                              action='store_false', 
                                              default=None, 
                        help='desactivate the filtering of Phi-X174 sequences')
    
    parser.add_argument('--no-seeds',            dest ='orgasm:noseeds', 
                                              metavar='no-seeds', 
                                              action='append',
                                              default=[], 
                                              type=str, 
                        help='protein or nucleic probes that will be used to counterselect seeds; ' 
                        'either a fasta file containing '
                        'probe sequences or the name of one of the internal set of seeds '
                        'among %s' % str(list(filter(lambda s: s.startswith('prot') or 
                                                s.startswith('nuc'),dir(orgasm.samples)))))

    parser.add_argument('--kup',              dest='orgasm:kup', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='The word size used to identify the seed reads '
                             '[default: protein=4, DNA=12]')

    parser.add_argument('--identity',         dest='orgasm:identity', 
                                              type=float, 
                                              action='store', 
                                              default=0.5, 
                        help='The fraction of word'
                             '[default: 0.5]')

    parser.add_argument('--clean',    dest='buildgraph:clean', 
                                              action='store_true', 
                                              default=None, 
                        help='Erase previously existing graph to restart a new assembling')
    
    parser.add_argument('--force-seeds',    dest='buildgraph:forceseeds', 
                                              action='store_true', 
                                              default=None, 
                        help='Force to reuse all the seeds')
    


def estimateMinRead(index,minoverlap,coverage):
    MINREAD=10
    MINOVERLAP=50
    minread =  (index.getReadSize() - minoverlap) * coverage / index.getReadSize()
    if minread < MINREAD:
        minoverlap = index.getReadSize() - (MINREAD * index.getReadSize() / coverage)
        minread = MINREAD
    if  minoverlap< MINOVERLAP:
        minread =  MINREAD
        minoverlap = MINOVERLAP
    return minread,minoverlap


def run(config):
    
    logger=config['orgasm']['logger']
    progress = config['orgasm']['progress']
    output = getOutput(config) 
    lowfilter=not config['buildgraph']['lowcomplexity']
    coverageset=config['buildgraph']['coverage'] is not None
    snp=config['buildgraph']['snp']
    assmax = config['buildgraph']['assmax']*2
    covratio = config['buildgraph']['covratio']
    covratiofill = config['buildgraph']['covratiofill']
    mincov = config['buildgraph']['mincov']

        
            

    logger.info("Building De Bruijn Graph")


    minoverlap = config['buildgraph']['minoverlap']
    logger.info('Minimum overlap between read: %d' % minoverlap)

    r = getIndex(config)
    adapterSeq3, adapterSeq5 = getAdapters(config)
    
    ecoverage,x,newprobes = getSeeds(r,config)   
            
    # Force the coverage to the specified value
    
    if coverageset:
        coverage = config['buildgraph']['coverage']
    else:
        coverage = ecoverage
    
    mincoverage = coverage // covratio

    if config['buildgraph']['clean'] or not os.path.exists("%s.oax" % output):
        
        #
        #  This is a new assembling or at least we are cleaning previous assembling
        # 
        
        newprobes=None

        ##
        if os.path.exists("%s.oax" % output):
            logger.info('Cleaning previous assembling')
            os.remove("%s.oax" % output)
        else:
            logger.info('No previous assembling')

        logger.info("Starting a new assembling")
        
        # Create the assembler object
        asm = Assembler(r)    
        logger.info('Coverage estimated from probe matches at : %d' % ecoverage)
        
        if coverageset:
            logger.info('Coverage forced by user at : %d' % coverage)
            minread,minoverlap = estimateMinRead(r,config['buildgraph']['minoverlap'],
                                                 mincoverage)
            if config['buildgraph']['minread'] is not None:
                minread = config['buildgraph']['minread']
            
        else:
        ##########################
        #
        # If minread is not specified we initiate the assembling
        # based on the coverage estimated from protein match
        # to obtain a better coverage estimation
        #
        ##########################
            if config['buildgraph']['minread'] is None:
    
                minread,minoverlap = estimateMinRead(r,config['buildgraph']['minoverlap'],
                                                     mincoverage)
                minread//=4
                if minread<5:
                    minread=5
        
                logger.info('Assembling of %d pb for estimating actual coverage' % config['buildgraph']['testrun'])
                
                s = matchtoseed(x,r,newprobes)
                
                #logger.info("minread = %d mincov = %d coverage = %d %d" % (minread,mincov,coverage,minoverlap))

                # Run the first assembling pass
                a = tango(asm,s,mincov=mincov,       # @UnusedVariable
                                minread=minread,
                                minoverlap=minoverlap,
                                lowfilter=lowfilter,
                                adapters3=adapterSeq3,
                                adapters5=adapterSeq5,
                                maxjump=0, 
                                cycle=1,
                                nodeLimit=config['buildgraph']['testrun'] * 2,
                                progress=progress,
                                logger=logger)
            
                # Clean small unsuccessful extensions
                asm.cleanDeadBranches(maxlength=10)
                
                # and too low covered assembling
                if coverageset:
                    cutLowCoverage(asm,int(coverage),terminal=True)
                else:
                    cutLowCoverage(asm,int(mincoverage),terminal=True)
                    
                if snp:
                    cutSNPs(asm)
                
                if config['buildgraph']['smallbranches'] is not None:
                    smallbranches = config['buildgraph']['smallbranches']
                else:
                    smallbranches = estimateDeadBrancheLength(asm)
                    
                logger.info("Dead branch length set to : %d bp" % smallbranches)
                
                asm.cleanDeadBranches(maxlength=smallbranches)
                
                if len(asm) > 0:
                    score,length,ecoverage = coverageEstimate(asm,x,r)  # @UnusedVariable
                    if not coverageset:
                        coverage = ecoverage    
                    
                    mincoverage = coverage // covratio

                    minread,minoverlap = estimateMinRead(r,config['buildgraph']['minoverlap'],
                                                         mincoverage) 
                    logger.info("coverage estimated : %dx based on %d bp (minread: %d)" %(coverage,length/2,minread))
                else:
                    logger.error('Nothing assembled - Assembling aborted')
                    sys.exit(1)
                                
                # Create the assembler object
                asm = Assembler(r)
            else:
                minread=config['buildgraph']['minread']


    else:
        
        #
        #  We are extending a previous assembling
        # 

        logger.info('Restoring previous assembling')
        asm = restoreGraph("%s.oax" % output,r,x)
        logger.info("Graph size %d bp" % int(len(asm)/2))
        logger.info("Entering in fill gaps mode")
                
        # Clean small unsuccessful extensions
        asm.cleanDeadBranches(maxlength=10)
        
        if len(asm) == 0:
            logger.error('The assembling is empty - Stop the assembling process')
            sys.exit(1)
    
        # estimate coverage from the graph
        score,length,ecoverage = coverageEstimate(asm,x,r)  # @UnusedVariable
    
        # If no coverage specified coverage = estimated coverage
        if not coverageset:
            coverage = ecoverage 

        mincoverage = coverage // covratio

        if not coverageset:
            cutLowCoverage(asm,int(mincoverage),terminal=True)
        else:
            cutLowCoverage(asm,int(coverage),terminal=True)

        if config['buildgraph']['smallbranches'] is not None:
            smallbranches = config['buildgraph']['smallbranches']
            logger.info("     Dead branch length forced by user to : %d bp" % smallbranches)
        else:
            smallbranches = estimateDeadBrancheLength(asm)
            logger.info("     Dead branch length set to : %d bp" % smallbranches)
    
        asm.cleanDeadBranches(maxlength=smallbranches)
    
        if len(asm) == 0:
            logger.error('The assembling is empty - Stop the assembling process')
            sys.exit(1)
    
        # reestimate coverage
        score,length,ecoverage = coverageEstimate(asm,x,r)  # @UnusedVariable
        
        logger.info('Coverage estimated from the assembling graph : %d' % ecoverage)
        
        if not coverageset:
            coverage = ecoverage  
            
        mincoverage = coverage // covratio

        # cleanup snp bubble in the graph    
        if snp:
            cutSNPs(asm)
            
        # according to the minread option estimate it from coverage or use the specified value
    
        if config['buildgraph']['minread'] is None:
            minread,minoverlap = estimateMinRead(r,config['buildgraph']['minoverlap'],
                                                 mincoverage)
    
            if minread<5:
                minread=5
        else:
            minread,minoverlap = estimateMinRead(r,config['buildgraph']['minoverlap'],
                                                 mincoverage)
            minread=config['buildgraph']['minread']

                    
        cg = asm.compactAssembling(verbose=False)
        genesincontig(cg,r,x)
        
        meanlength,sdlength = estimateFragmentLength(asm)
        
        if config['orgasm']['back'] is not None:
            back = config['orgasm']['back']
        elif config['orgasm']['back'] is None and meanlength is not None:
            back = int(meanlength + 4 * sdlength)
            if back > 500:
                back=500
        else:
            back = 300

        
        scaffold(asm,cg,
                 minlink=5,
                 back=back,
                 addConnectedLink=False,
                 logger=logger)
        
        with open(output+'.intermediate.gml','w') as gmlfile:
            print(cg.gml(),file=gmlfile)
                
    logger.info("based on a coverage of : %d minread set to: %d)" %(coverage,minread))
                
    # Convert matches in seed list
    s = matchtoseed(x,r,newprobes)
    
    if s:
        #############################################
        #
        # We now run the main assembling process
        #
        #############################################
        
        resetusedreads()
    
        logger.info('Starting the assembling')
        # logger.info("minread = %d mincov = %d coverage = %d %d" % (minread,mincov,coverage,minoverlap))
        
        # Run the first assembling pass
        a = tango(asm,s,mincov=mincov,       #@UnusedVariable
                        minread=minread,
                        minoverlap=minoverlap,
                        lowfilter=lowfilter,
                        adapters3=adapterSeq3,
                        adapters5=adapterSeq5,
                        maxjump=0, 
                        cycle=1,
                        nodeLimit=assmax,
                        progress=progress,
                        logger=logger)
    
        # Clean small unsuccessful extensions
        asm.cleanDeadBranches(maxlength=10)
            
        # and too low covered assembling
        if coverageset:
            cutLowCoverage(asm,int(coverage),terminal=True)
        else:
            cutLowCoverage(asm,int(mincoverage),terminal=True)
            
        
        # cleanup snp bubble in the graph    
        if snp:
            cutSNPs(asm)
        
        if config['buildgraph']['smallbranches'] is not None:
            smallbranches = config['buildgraph']['smallbranches']
        else:
            smallbranches = estimateDeadBrancheLength(asm)
            logger.info("     Dead branch length setup to : %d bp" % smallbranches)
    
        asm.cleanDeadBranches(maxlength=smallbranches)
    
    
        # reestimate coverage
        
        if len(asm) == 0:
            logger.error('The assembling is empty - Stop the assembling process')
            sys.exit(1)
    
        score,length,ecoverage = coverageEstimate(asm,x,r)  # @UnusedVariable
        
        if not coverageset:
            coverage = ecoverage  
            
        mincoverage = coverage // covratio
        
    #     if coverage < 30:
    #         sys.exit()
    
   
        logger.info("coverage estimated : %d based on %d bp" %(coverage,length/2))
            
        meanlength,sdlength = estimateFragmentLength(asm)
        
        if config['orgasm']['back'] is not None:
            back = config['orgasm']['back']
        elif config['orgasm']['back'] is None and meanlength is not None:
            back = int(meanlength + 4 * sdlength)
            if back > 500:
                back=500
        else:
            back = 300
            
        if meanlength is not None:
            logger.info("Fragment length estimated : %f pb (sd: %f)" % (meanlength,sdlength))
    
        cg = asm.compactAssembling(verbose=False)
        genesincontig(cg,r,x)
        scaffold(asm,
                 cg,
                 minlink=5,
                 back=back,
                 addConnectedLink=False,
                 logger=logger)
        
        with open(output+'.intermediate.gml','w') as gmlfile:
            print(cg.gml(),file=gmlfile)

    if len(asm) == 0:
        logger.error('The assembling is empty - Stop the assembling process')
        sys.exit(1)

 
    ###################################################
    #
    # We now fill the gaps between the contigs
    #
    ###################################################
    
    mincoverage = coverage // covratiofill

    if config['buildgraph']['minread'] is None:
        logger.info("Estimating minread")
        minread,minoverlap = estimateMinRead(r,config['buildgraph']['minoverlap'],
                                             mincoverage)
        if minread<5:
            minread=5

    logger.info("minread set to: %d" % minread)
 
    delta = 1
    maxfillgaps=config['buildgraph']['maxfillgaps']
    fillcycle=0
    # Run the fill gap procedure    
    while  (delta > 0 or delta < -100) and (maxfillgaps < 0 or fillcycle < maxfillgaps ):
        fillcycle+=1
        
        # intermediate graph are saved before each gap filling step
        dumpGraph(output+'.intermediate.oax',asm)
        
        delta = fillGaps(asm,back=back,
                       minread=minread,
                       maxjump=0,
                       minoverlap=minoverlap,
                       cmincov=1,
                       emincov=mincoverage,
                       gmincov=mincoverage,
                       lowfilter=lowfilter,
                       adapters5 = adapterSeq5,
                       adapters3 = adapterSeq3,
                       snp=snp,
                       nodeLimit=assmax)

        print('',file=sys.stderr)
        print('',file=sys.stderr)
        print('',file=sys.stderr)
        print('==================================================================',file=sys.stderr)
        print('',file=sys.stderr)
        
        cg = asm.compactAssembling(verbose=False)
        genesincontig(cg,r,x)
        scaffold(asm,cg,minlink=5,back=back,addConnectedLink=False,
                 logger=logger)
        with open(output+'.intermediate.gml','w') as gmlfile:
            print(cg.gml(),file=gmlfile)
        
        if meanlength is None:
            meanlength,sdlength = estimateFragmentLength(asm)
            if config['orgasm']['back'] is None and meanlength is not None:
                logger.info("Fragment length estimated : %f pb (sd: %f)" % (meanlength,sdlength))
                back = int(meanlength + 4 * sdlength)  
                if back > 500:
                    back=500
                         
        print('',file=sys.stderr)
        print('==================================================================',file=sys.stderr)
        print('',file=sys.stderr)
        
    ###################################################
    #
    # Finishing of the assembling
    #
    ###################################################

    if snp:
        logger.info("Clean polymorphisms")
        cutSNPs(asm)
        
    asi = len(asm)+1
    
    logger.info("Clean dead branches")
    while (asi>len(asm)):
        asi=len(asm)
        smallbranches = estimateDeadBrancheLength(asm)
        logger.info("     Dead branch length setup to : %d bp" % smallbranches)
        asm.cleanDeadBranches(maxlength=smallbranches)
        
    cg = asm.compactAssembling(verbose=False)
    
    
    if len(asm) == 0:
        logger.error('The assembling is empty - Stop the assembling process')
        sys.exit(1)

    score,length,ecoverage = coverageEstimate(asm,x,r)  # @UnusedVariable
    if not coverageset:
        coverage=ecoverage

    if snp:
        logger.info("Clean polymorphisms phase 2")
        cutSNPs(asm)
        
    logger.info("Clean low coverage terminal branches")
    if coverageset:
        cutLowCoverage(asm,int(coverage),terminal=False)
    else:
        cutLowCoverage(asm,int(mincoverage),terminal=True)
        logger.info("Clean low coverage internal branches")
        cutLowCoverage(asm,int(mincoverage),terminal=False)
        
    logger.info("Saving the assembling graph")
    dumpGraph(output+'.oax',asm)
    asm = restoreGraph(output+'.oax',r,x)
        
    cg = asm.compactAssembling(verbose=False)     
        
    logger.info("Scaffold the assembly")
    scaffold(asm,cg,minlink=5,back=int(back),addConnectedLink=False,
                 logger=logger)
    genesincontig(cg,r,x)
    with open(output+'.gml','w') as gmlfile:
        print(cg.gml(),file=gmlfile)



'''
Created on 20 oct. 2012

@author: coissac
'''

from distutils.command.build import build as ori_build
from obidistutils.serenity.checksystem import is_mac_system


class build(ori_build):
    
    def has_ext_modules(self):
        return self.distribution.has_ext_modules()
    
    def has_pidname(self):
        return is_mac_system()

    def has_doc(self):
        return True
    
    def has_littlebigman(self):
        return True
    
    try:
        from obidistutils.command.build_sphinx import build_sphinx  # @UnusedImport

        sub_commands = [("littlebigman",has_littlebigman),
                         ('pidname',has_pidname)
                       ] \
                       + ori_build.sub_commands + \
                       [('build_sphinx',has_doc)]
    except ImportError:
        sub_commands = [("littlebigman",has_littlebigman),
                         ('pidname',has_pidname)
                       ] \
                       + ori_build.sub_commands
        



#cython: language_level=3

'''
Created on 27 mars 2016

@author: coissac
'''

import argparse
import sys


from .command import getCommandsList

class ObiParser(argparse.ArgumentParser): 
    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2)

cpdef buildArgumentParser(str configname, 
                          str softname):
    parser = ObiParser()
    
    parser.add_argument('--version',   dest='%s:version' % configname, 
                                       action='store_true', 
                                       default=False, 
                        help='Print the version of %s' % softname)

    parser.add_argument('--log',       dest='%s:log' % configname, 
                                       action='store',
                                       type=str,
                                       default=None, 
                        help='Create a logfile')

    parser.add_argument('--no-progress', dest='%s:progress' % configname, 
                                       action='store_false', 
                                       default=None, 
                        help='Do not print the progress bar during analyzes')

    subparsers = parser.add_subparsers(title='subcommands',
                                       description='valid subcommands',
                                       help='additional help')
    
    commands = getCommandsList()
    
    for c in commands:
        module = commands[c]
        
        if hasattr(module, "run"):
            if hasattr(module, "__title__"):
                sub = subparsers.add_parser(c,help=module.__title__)
            else:
                sub = subparsers.add_parser(c)
    
            if hasattr(module, "addOptions"):
                module.addOptions(sub)
            
            sub.set_defaults(**{'%s:module'  % configname : module})
                        
    return parser



# cython: language_level=3

from orgasm.graph._graph cimport *
from orgasm.indexer._orgasm cimport *

cdef class AsmbGraph(DiGraph):
    cdef Index _readIndex
    
    cpdef dict addNode(self,int node)
    cpdef dict addEdge(self,int node1, int node2)
    #def dict getNodeAttr(self, int node)




'''
Created on 28 sept. 2014

@author: coissac
'''
from orgasm import getOutput,getIndex, getSeeds
from orgasm.tango import restoreGraph, estimateFragmentLength, genesincontig,\
    scaffold, cutLowCoverage, estimateDeadBrancheLength, dumpGraph, cutLowSeeds
import sys

__title__="Cut low coverage edge in an assembling graph"

default_config = { 'coverage'      : None,
                   'smallbranches' : None,
                   'lowseeds'      : None
                 }

def addOptions(parser):
    parser.add_argument(dest='orgasm:indexfilename',  metavar='index', 
                        help='index root filename (produced by the oa index command)')
    
    parser.add_argument(dest='orgasm:outputfilename',     metavar='output', 
                                                          nargs='?', 
                                                          default=None,
                        help='output prefix' )
    
    
    parser.add_argument('--coverage',         dest='cutlow:coverage', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='All edges with a coverage below this value will be deleted')
    
    parser.add_argument('--seeds',         dest='cutlow:lowseeds', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='All edges with a coverage below this value will be deleted')
    
    parser.add_argument('--smallbranches',    dest='cutlow:smallbranches', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='maximum length of the branches to cut during '
                             'the cleaning process [default: <estimated>]')

    parser.add_argument('--back',             dest='orgasm:back', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='the number of bases taken at the end of '
                             'contigs to jump with pared-ends [default: <estimated>]')
    


def run(config):

    logger=config['orgasm']['logger']
    output = getOutput(config)
    
    if (config['cutlow']['coverage'] is None and
        config['cutlow']['lowseeds'] is None
       ):
        logger.error("No threshold specified")
        sys.exit(1)

    r = getIndex(config)
    xxx,x,newprobes = getSeeds(r,config)  
    
    asm = restoreGraph(output+'.oax',r,x)

    logger.info("Evaluate fragment length")
    
    meanlength,sdlength = estimateFragmentLength(asm)
    
    if meanlength is not None:
        logger.info("Fragment length estimated : %f pb (sd: %f)" % (meanlength,sdlength))

    if config['orgasm']['back'] is not None:
        back = config['orgasm']['back']
    elif config['orgasm']['back'] is None and meanlength is not None:
        back = int(meanlength + 4 * sdlength)
        if back > 500:
            back=500
    else:
        back = 300
        
    logger.info("Cut low coverage")

    if config['cutlow']['coverage'] is not None:
        cutLowCoverage(asm,config['cutlow']['coverage'],terminal=False)
    
    if config['cutlow']['lowseeds'] is not None:
        cutLowSeeds(asm,config['cutlow']['lowseeds'],x,terminal=False)
    
    if config['cutlow']['smallbranches'] is not None:
        smallbranches = config['cutlow']['smallbranches']
    else:
        smallbranches = estimateDeadBrancheLength(asm)
        logger.info("Dead branch length setup to : %d bp" % smallbranches)
    
    asm.cleanDeadBranches(maxlength=smallbranches)
    
    cg = asm.compactAssembling(verbose=False)
    genesincontig(cg,r,x)

    scaffold(asm,cg,minlink=5,back=int(back),addConnectedLink=False,
                 logger=logger)
     
    with open(output+'.gml','w') as gmlfile:
        print(cg.gml(),file=gmlfile)

    dumpGraph(output+'.oax',asm)

              



# cython: language_level=3

from cpython.ref cimport PyObject,Py_XDECREF

#
# C API to Bytes object
#

from cpython.bytes  cimport PyBytes_Size, PyBytes_FromStringAndSize

# We explicitly want access to the raw version of the PyBytes_AS_STRING macro
# that can be called with a PyObject* 
cdef extern from "Python.h":
    char* PyBytes_AS_STRING(PyObject* obj)

#
# C API to Dict object
#

from cpython.dict  cimport PyDict_New,PyDict_GetItemString,PyDict_SetItemString
from cpython.dict  cimport PyDict_Next,PyDict_SetItem,PyDict_GetItem,PyDict_Size

#
# C API to List object
#

from cpython.list  cimport PyList_GET_SIZE, PyList_Append

#
# C API to Tuple object
#

from cpython.tuple  cimport PyTuple_GET_ITEM

#
# C API to int object
#

from cpython.int  cimport PyInt_AsSsize_t,PyInt_FromLong

#
# C API to the libc
#

from libc.stdlib  cimport malloc,realloc,free
from libc.math cimport log
from libc.string cimport strlen,memcpy

cdef extern from "strings.h":
    void bzero(void *s, size_t n)

from orgasm.graph._graphmultiedge cimport DiGraphMultiEdge
from orgasm.indexer._orgasm cimport Index

cdef dict listCodons2Dict(tuple codons)
cpdef dict bindCodons(dict codon1, dict codon2)
cpdef dict codon(char aa, bint direct=?)

cdef struct protmatch:
    int        protid
    int        position
    protmatch* next

cdef struct dnastate:
#    dnastate*  error
    protmatch* match
    dnastate*  a 
    dnastate*  c 
    dnastate*  t 
    dnastate*  g 
   
cdef extern from "inttypes.h":
    ctypedef int int32_t  
    ctypedef int int64_t  
    ctypedef unsigned int uint32_t
    ctypedef unsigned int uint64_t

cdef extern from "orgasm.h":
    uint32_t expanded8bitsnuc[]
 

cdef class AhoCorasick:

    cdef dnastate *_states
    cdef size_t   _statesize
    cdef size_t   _statecount
    
    cdef protmatch *_matches 
    cdef size_t    _matchsize
    cdef size_t    _matchcount
    
    # the automata has been finalized
    cdef bint      _finalized
    
    cdef size_t    _step
    cdef size_t    _depth
    cdef int       _maxseqid
    cdef size_t    _maxpos
    
    cdef dict      _seqid
    cdef dict      _rseqid
    

    cdef dnastate *getNextState(self,dnastate *base, int letter)
    cdef protmatch* setAsTerminal(self, dnastate* base, int protid, int position)
    cdef dnastate* simpleMatch(self,char* cword, size_t lword)
    cdef dnastate* longestSuffix(self,char* cword, size_t lword)
    cdef setErrorLink(self)
    cpdef finalize(self)
    cdef int addCWord(self,char* cword, size_t lword, int protid, int position)
    cdef int addAutomata(self,dict automata,int protid, int position)
    cpdef list match(self,char* sequence)    
    cpdef DiGraphMultiEdge asGraph(self)
    
cdef class ProtAhoCorasick(AhoCorasick):

    cpdef int addSequence(self,bytes sequence, object seqid=?, size_t kup=?)
    cpdef scanIndex(self,Index seqindex,int minmatch=?, int maxmatch=?, int covmin=?, bint progress=?)

cdef class NucAhoCorasick(AhoCorasick):

    cpdef int addSequence(self,bytes sequence, object seqid=?, size_t kup=?)
    cpdef scanIndex(self,Index seqindex,int minmatch=?, int maxmatch=?, int covmin=?, bint progress=?)

 
 


# cython: language_level=3

from ._fibominintheap cimport *

import sys

cdef int cmpMinIntFunc (const_FiboNode_ptr_const node1, const_FiboNode_ptr_const node2):
    cdef IntData *data1 = <IntData*> node1
    cdef IntData *data2 = <IntData*> node2

    return -1 if (data1.priority < data2.priority) else 1


cdef class FibonacciMinIntHeap(FibonacciIntHeap):

    def __init__(self,object data=None,size_t chunksize=1000):
        
        cdef size_t i
        cdef size_t priority 
        cdef int32_t di
          
        if fiboTreeInit(&self._treedat, cmpMinIntFunc) != 0:
            raise SystemError
        
        self._last=NULL
        self._size=0
        self._chunksize=chunksize
        self._arenacount=0
        self._realloc()
        self._nodes = {}
               
        if data is not None:
            for d in data:
                try:
                    priority = d[0]
                    di = d[1]
                except :
                    priority = d
                    di = priority
                    
                self.push(priority,di)
                
        
    cpdef tuple min(self):
        cdef FiboNode *node 
        cdef IntData     *d
        
        assert self._size > 0,"Heap is empty"

        node = fiboTreeMin(&self._treedat)
        d = <IntData*>node
             
        return d.data,d.priority
    
    cpdef int32_t minInt(self):
        cdef FiboNode *node 
        cdef IntData     *d
        
        assert self._size > 0,"Heap is empty"

        node = fiboTreeMin(&self._treedat)
        d = <IntData*>node
             
        return d.data
    
    
        



# cython: language_level=3

from orgasm.tango import pairEndedConnected
from orgasm.tango import estimateFragmentLength
from orgasm.tango import logOrPrint

from orgasm.utils import bytes2str


def buildReadSequence(self,assgraph,identifier,path,back=200,nlength=20,logger=None):
    
    #
    # Build a dictionary with:
    #      - Keys   = stemid
    #      - Values = edge descriptor (from,to,x)
    #
    readsize = self.index.getReadSize()
    frglen,frglensd = estimateFragmentLength(self) # @UnusedVariable
    
    alledges = dict((assgraph.getEdgeAttr(*e)['stemid'],e) 
                    for e in assgraph.edgeIterator(edgePredicate = lambda i: 'stemid' in assgraph.getEdgeAttr(*i)))

    seq=[]
    oldstem=None
    oldid=None
    oldstemclass=None
     
    rank = 1
    forceconnection=False
    
    for stemid in path:
        if stemid != 0:
            stem              = alledges[stemid]
            attr              = assgraph.getEdgeAttr(*stem)
            stemclass         = attr['class']
            sequence          = attr['path']
            
            if rank==1 and stemclass!="sequence":
                raise RuntimeError("A path cannot start on a gap")

                                                    # Switch the stem to a dashed style
            graphics          = attr.get('graphics',{})
            attr['graphics']  = graphics
            graphics['style'] = 'dashed'
            
            # Manage step rank information of each step
            allsteps = attr.get('steps',{})
            steps = allsteps.get(identifier,[])
            steps.append(rank)
            allsteps[identifier]=steps
            attr['steps']=allsteps
            
            if oldstem is not None:
                connected,ml,sl,delta = pairEndedConnected(self,assgraph,oldid,stemid,back)  # @UnusedVariable
                if oldstem[1]==stem[0]:
                    if oldstemclass=="sequence":
                        if stemclass=="sequence":                   # Link between 2 sequences
                            if ml is not None:
                                logOrPrint(logger,
                                           "Both segments %d and %d are connected (paired-end=%d frg length=%f sd=%f)" % 
                                           (oldid,stemid,connected,float(ml),float(sl)))
            
                            else:
                                logOrPrint(logger,
                                           "Both segments %d and %d are connected but covered by 0 paired-end" % 
                                           (oldid,stemid))
            
                                                            
                        elif stemclass[0:9]=="scaffold:":           # Link a sequence and a gap
                            logOrPrint(logger,
                                        "Both segments %d and %d are disconnected" % attr['link'])
    
                            if stemclass=="scaffold:paired-end":
                                logOrPrint(logger,
                                           "   But linked by %d pair ended links (gap length=%f sd=%f)" % 
                                           (attr['pairendlink'],
                                            attr['length'],
                                            attr['gapsd']))
                                           
    
                            elif stemclass=="scaffold:forced":
                                logOrPrint(logger,"   Connection is forced")
                                
                                if attr['pairendlink'] > 0:
                                    logOrPrint(logger,
                                               "   But asserted by %d pair ended links (gap length=%f sd=%f)" % 
                                               (attr['pairendlink'],
                                                attr['length'],
                                                attr['gapsd']))
                            elif stemclass=="scaffold:overlap":
                                logOrPrint(logger,
                                           "   But overlap by %dbp supported by %d pair ended links" % 
                                           (-attr['length'],
                                            attr['pairendlink']))
                                        
                    elif oldstemclass[0:9]=="scaffold:" and stemclass[0:9]=="scaffold:":
                        raise RuntimeError('A scaffold link must be followed by a sequence %d --> %d' %
                                           (oldid,stemid))           
                        
                elif forceconnection:
                    logOrPrint(logger,"   Connection is forced")
                    if connected > 0:
                        glength = int(frglen-ml - readsize) 

                        logOrPrint(logger,
                                   "   But asserted by %d pair ended links (gap length=%f sd=%f)" % 
                                   (connected,
                                    glength,
                                    sl))

                    else:
                        
                        logOrPrint(logger,
                                   "Without any support from pair ended links")
                        glength =  nlength
                    
                    seq.append([0] * int(glength) + [stem[0]])

                    # Add the foced link to the compact assembly graph    
                    flink = assgraph.addEdge(oldstem[1],stem[0])
                    rlink = assgraph.addEdge(-stem[0],-oldstem[1])
                    flink['label']="Forced (%d)  %d -> %d" % (connected,oldid,stemid)
                    flink['graphics']={'width':1,
                                       'arrow':'last',
                                       'fill':'0xFF0000',
                                       'style':'dashed'

                                      }
                    rlink['label']="Forced (%d)  %d -> %d" % (connected,-stemid,-oldid)
                    rlink['graphics']={'width':1,
                                       'arrow':'last',
                                       'fill':'0xFF0000',
                                       'style':'dashed'
                                      }
                else:
                    raise AssertionError('Disconnected path between stem '
                                         '%d and %d only %d pair ended links' % (oldid,stemid,connected))



            if stemclass=="sequence":
                if attr['length'] > 10:
                    attr['label']="%d : %s->(%d)->%s  [%d] @ %s" % (stemid,attr['sequence'][0:5].decode('ascii'),
                                                                    attr['length'],
                                                                    attr['sequence'][-5:].decode('ascii'),
                                                                    int(attr['weight']),
                                                                    attr['steps'])
                else:
                    attr['label']="%d : %s->(%d)  [%d] @ %s" % (stemid,
                                                                attr['sequence'].decode('ascii'),
                                                                attr['length'],
                                                                int(attr['weight']),
                                                                attr['steps'])
                    
 #           if oldstemclass is not None:
            if seq:
                seq.append(sequence[1:])
            else:
                seq.append(sequence)

            rank+=1
            oldstem = stem
            oldid=stemid
            oldstemclass=stemclass
            
            forceconnection=False
        else:
            forceconnection=True

    
                        
            
        
    s1 = alledges[path[-1]]
    s2 = alledges[path[0]]
    sid1=assgraph.getEdgeAttr(*s1)['stemid']
    sid2=assgraph.getEdgeAttr(*s2)['stemid']
    sclass2=assgraph.getEdgeAttr(*s2)['class']
    connected,ml,sl,delta = pairEndedConnected(self,            # @UnusedVariable
                                               assgraph,
                                               sid1,
                                               sid2,
                                               back)  
    
    if s1[1]==s2[0]:
        logOrPrint(logger, "Path is circular")
                    
            
        circular=True
        seq[-1].pop()
    else:
        if sclass2!="sequence":
            raise RuntimeError("A path cannot ends on a gap")
        
        if forceconnection:
            logOrPrint(logger,"Circular connection forced",)
            logOrPrint(logger,"Linked by %d pair ended links" %  connected)
                
            seq.append([0]*int(nlength))
            circular=True
        else:
            logOrPrint(logger,"Path is linear")
            circular=False

#        seq.insert(0,self.index.getRead(s2[0],0,self.index.getReadSize()).lower())

    sequence=[]
    for l in seq:
        sequence.extend(l)
         
    readsize=self.index.getReadSize()
        
    if circular:
        sequence = sequence[readsize:] + sequence[0:readsize]
         
    
    return sequence

def path2sam(self,assgraph,path,identifier="contig",minlink=10,nlength=20,back=200,logger=None,tags=[]):
    '''
    Convert a path in an compact assembling graph in a fasta formated sequences.
    
    :param assgraph: The compact assembling graph as produced by the
                     :py:meth:`~orgasm.assembler.Assembler.compactAssembling` method
    :type assgraph:  :py:class:`~orgasm.graph.DiGraphMultiEdge`
    :param path:     an ``iterable`` providing an ordered list of ``stemid`` indicating
                     the path to follow.  
    :type path:      an ``iterable`` over :py:class:`int`
    :param identifier: the identifier used in the header of the fasta formated sequence
    :type identifier:  :py:class:`bytes`
    :param minlink:  the minimum count of pair ended link to consider 
                     for asserting the relationship
    :type minlink:   :py:class:`int`
    :param nlength:  how many ``N`` must be added between two segment of sequences only connected
                     by pair ended links
    :type nlength:   :py:class:`int`
    :param back:     How many base pairs must be considered at the end of each edge
    :type back:      :py:class:`int`
    
    :returns: a string containing the fasta formated sequence
    :rtype: :py:class:`bytes`

    :raises: :py:class:`AssertionError`
    '''
    
    readsize = self.index.getReadSize()
#    frglen,frglensd = estimateFragmentLength(self) # @UnusedVariable
        
    sequence = buildReadSequence(self,assgraph,identifier,path,back,nlength,logger)
#     seq=[]
#     oldstem=None
#     oldid=None
#     oldstemclass=None
#      
#     rank = 1
#     forceconnection=False
#     
#     for stemid in path:
#         if stemid != 0:
#             stem              = alledges[stemid]
#             attr              = assgraph.getEdgeAttr(*stem)
#             stemclass         = attr['class']
#             sequence          = attr['path']
#             
#             if rank==1 and stemclass!="sequence":
#                 raise RuntimeError("A path cannot start on a gap")
# 
#                                                     # Switch the stem to a dashed style
#             graphics          = attr.get('graphics',{})
#             attr['graphics']  = graphics
#             graphics['style'] = 'dashed'
#             
#             # Manage step rank information of each step
#             allsteps = attr.get('steps',{})
#             steps = allsteps.get(identifier,[])
#             steps.append(rank)
#             allsteps[identifier]=steps
#             attr['steps']=allsteps
#             
#             if oldstem is not None:
#                 connected,ml,sl,delta = pairEndedConnected(self,assgraph,oldid,stemid,back)  # @UnusedVariable
#                 if oldstem[1]==stem[0]:
#                     if oldstemclass=="sequence":
#                         if stemclass=="sequence":                   # Link between 2 sequences
#                             if ml is not None:
#                                 logOrPrint(logger,
#                                            "Both segments %d and %d are connected (paired-end=%d frg length=%f sd=%f)" % 
#                                            (oldid,stemid,connected,float(ml),float(sl)))
#             
#                             else:
#                                 logOrPrint(logger,
#                                            "Both segments %d and %d are connected but covered by 0 paired-end" % 
#                                            (oldid,stemid))
#             
#                                                             
#                         elif stemclass[0:9]=="scaffold:":           # Link a sequence and a gap
#                             logOrPrint(logger,
#                                         "Both segments %d and %d are disconnected" % attr['link'])
#     
#                             if stemclass=="scaffold:paired-end":
#                                 logOrPrint(logger,
#                                            "   But linked by %d pair ended links (gap length=%f sd=%f)" % 
#                                            (attr['pairendlink'],
#                                             attr['length'],
#                                             attr['gapsd']))
#                                            
#     
#                             elif stemclass=="scaffold:forced":
#                                 logOrPrint(logger,"   Connection is forced")
#                                 
#                                 if attr['pairendlink'] > 0:
#                                     logOrPrint(logger,
#                                                "   But asserted by %d pair ended links (gap length=%f sd=%f)" % 
#                                                (attr['pairendlink'],
#                                                 attr['length'],
#                                                 attr['gapsd']))
#                             elif stemclass=="scaffold:overlap":
#                                 logOrPrint(logger,
#                                            "   But overlap by %dbp supported by %d pair ended links" % 
#                                            (-attr['length'],
#                                             attr['pairendlink']))
#                                         
#                     elif oldstemclass[0:9]=="scaffold:" and stemclass[0:9]=="scaffold:":
#                         raise RuntimeError('A scaffold link must be followed by a sequence %d --> %d' %
#                                            (oldid,stemid))           
#                         
#                 elif forceconnection:
#                     logOrPrint(logger,"   Connection is forced")
#                     if connected > 0:
#                         glength = int(frglen-ml - self.index.getReadSize()) 
# 
#                         logOrPrint(logger,
#                                    "   But asserted by %d pair ended links (gap length=%f sd=%f)" % 
#                                    (connected,
#                                     glength,
#                                     sl))
# 
#                     else:
#                         
#                         logOrPrint(logger,
#                                    "Without any support from pair ended links")
#                         glength =  nlength
#                     
#                     seq.append([0] * int(glength) + [stem[0]])
# 
#                     # Add the foced link to the compact assembly graph    
#                     flink = assgraph.addEdge(oldstem[1],stem[0])
#                     rlink = assgraph.addEdge(-stem[0],-oldstem[1])
#                     flink['label']="Forced (%d)  %d -> %d" % (connected,oldid,stemid)
#                     flink['graphics']={'width':1,
#                                        'arrow':'last',
#                                        'fill':'0xFF0000',
#                                        'style':'dashed'
# 
#                                       }
#                     rlink['label']="Forced (%d)  %d -> %d" % (connected,-stemid,-oldid)
#                     rlink['graphics']={'width':1,
#                                        'arrow':'last',
#                                        'fill':'0xFF0000',
#                                        'style':'dashed'
#                                       }
#                 else:
#                     raise AssertionError('Disconnected path between stem '
#                                          '%d and %d only %d pair ended links' % (oldid,stemid,connected))
# 
# 
# 
#             if stemclass=="sequence":
#                 if attr['length'] > 10:
#                     attr['label']="%d : %s->(%d)->%s  [%d] @ %s" % (stemid,attr['sequence'][0:5].decode('ascii'),
#                                                                     attr['length'],
#                                                                     attr['sequence'][-5:].decode('ascii'),
#                                                                     int(attr['weight']),
#                                                                     attr['steps'])
#                 else:
#                     attr['label']="%d : %s->(%d)  [%d] @ %s" % (stemid,
#                                                                 attr['sequence'].decode('ascii'),
#                                                                 attr['length'],
#                                                                 int(attr['weight']),
#                                                                 attr['steps'])
#                     
#  #           if oldstemclass is not None:
#             if seq:
#                 seq.append(sequence[1:])
#             else:
#                 seq.append(sequence)
# 
#             rank+=1
#             oldstem = stem
#             oldid=stemid
#             oldstemclass=stemclass
#             
#             forceconnection=False
#         else:
#             forceconnection=True
# 
#     
#                         
#             
#         
#     s1 = alledges[path[-1]]
#     s2 = alledges[path[0]]
#     sid1=assgraph.getEdgeAttr(*s1)['stemid']
#     sid2=assgraph.getEdgeAttr(*s2)['stemid']
#     sclass2=assgraph.getEdgeAttr(*s2)['class']
#     connected,ml,sl,delta = pairEndedConnected(self,            # @UnusedVariable
#                                                assgraph,
#                                                sid1,
#                                                sid2,
#                                                back)  
#     
#     if s1[1]==s2[0]:
#         logOrPrint(logger, "Path is circular")
#                     
#             
#         circular=True
#         seq[-1].pop()
#     else:
#         if sclass2!="sequence":
#             raise RuntimeError("A path cannot ends on a gap")
#         
#         if forceconnection:
#             logOrPrint(logger,"Circular connection forced",)
#             logOrPrint(logger,"Linked by %d pair ended links" %  connected)
#                 
#             seq.append([0]*int(nlength))
#             circular=True
#         else:
#             logOrPrint(logger,"Path is linear")
#             circular=False
# 
# #        seq.insert(0,self.index.getRead(s2[0],0,self.index.getReadSize()).lower())
# 
#     sequence=[]
#     for l in seq:
#         sequence.extend(l)
#          
#     readsize=self.index.getReadSize()
#         
#     if circular:
#         sequence = sequence[readsize:] + sequence[0:readsize]
#          
#         
    reads = []
    location={}
    
    
    for i in range(len(sequence)):
        rid = sequence[i]
        rl={}
        reads.append(rl)       
#        if rid !=0 and not self.index.isFake(rid):
        if rid !=0:
            fake = self.index.isFake(rid)
            if not fake:
                allreads = self.index.getIds(rid)
                signe = 1 if rid > 0 else -1
                for rid in allreads[2]:
                    rid*=signe
                    prid = self.index.getPairedRead(rid)
                    rl[rid]=(prid,min(abs(rid),abs(prid)))
                    if rid not in location:
                        lp=[]
                        location[rid]=lp
                    else:
                        lp=location[rid]
                    lp.append(i+1)
            else:
                rl[rid]=(None,None)
                if rid not in location:
                    lp=[]
                    location[rid]=lp
                else:
                    lp=location[rid]
                lp.append(i+1)
                     
    sam=[]
    mapped=set()
    
    nreads = len(reads)
    
    def distreads(rid,a,b):
        if rid > 0:
            d = b - a + 1
        else:
            d = a - b
        if d < 0:
            d+=nreads
        return d
        
    
    for i in range(nreads):
        rlist=reads[i]
        for rid in rlist:
            flag=0
            
            paired,qname=rlist[rid]
            if paired is not None:
                npaired= self.index.getIds(paired)[0]

                paired_mapped = npaired in location
                flag |= 0x01            # template having multiple segments in sequencing
                                        # All our reads are paired   
                                        
                mapping=255  
                optional=""                   
    
            else:
                paired_mapped = False
                mapping=0
                optional="fk:i:1"
                
                
            readpos=i+1
            
            if paired_mapped:
                dist = [distreads(rid,readpos,x)
                        for x in location[npaired]]
                pairedpos=location[npaired][dist.index(min(dist))]
                if readpos < pairedpos:
                    tlen = pairedpos - readpos + readsize
                else:
                    tlen = -(readpos - pairedpos + readsize)
            else:
                qname = abs(rid)
                pairedpos=0
                tlen=0

            if paired_mapped:
                flag |= 0x02        # each segment properly aligned according to the aligner
                                    # paired read is part of the assembly        
                if rid > 0:      
                    flag |= 0x40    # the first segment in the template
                    flag |= 0x20    # SEQ of the next segment in the template being reverse complemented
                else:
                    flag |= 0x80    # the last segment in the template
            else:
                flag |= 0x08        # next segment in the template unmapped
                    
            
            if rid < 0:
                flag |= 0x10        # SEQ being reverse complemented
                
            if abs(rid) in mapped:
                flag|=0x100         # secondary alignment
            else:
                mapped.add(abs(rid))
                
            rseq = bytes2str(self.index.getRead(rid,
                                          0,
                                          self.index.getReadSize()).upper())
                 
            sam.append("R%09d\t%d\t%s\t%d\t%d\t%dM\t%s\t%d\t%d\t%s\t*" 
                                                % (qname,
                                                   flag,identifier,          
                                                   readpos,                           # Read position
                                                   mapping,                               # No mapping quality
                                                   readsize,                          # For the CIGAR string
                                                   '=' if paired_mapped else '*',     # 
                                                   pairedpos,
                                                   tlen,
                                                   rseq
                                                   ))
            
            if len(optional):
                sam[-1]=sam[-1]+ "\t" + optional
                
#    print(sequence[0:20])
#    print(reads[0:20])

    #sequence = b''.join(seq)
    #length = len(sequence)
    
    sam = "\n".join(sam)
    
    return sam



# cython: language_level=3

from ._fibointheap cimport *

cdef int cmpMinIntFunc(const_FiboNode_ptr_const node1, const_FiboNode_ptr_const node2)

cdef class FibonacciMinIntHeap(FibonacciIntHeap):


    cpdef tuple min(self)
    cpdef int32_t minInt(self)





#cython: language_level=3
from cpython.array cimport array

from orgasm.files.universalopener import uopen


def readFastq(filename):
    
    cdef str   line
    cdef bytes bline
    cdef bytes sid
    
    cdef int   seqid = 0
    cdef int   cut_slash
    cdef int   cut_space
    
    cdef array quality
    
    cdef bytes seq
    
    if isinstance(filename, str):
        filename = uopen(filename)
            
    for line in filename:
        bline = bytes(line,
                      encoding='ascii'
                     )
                
        while(bline[0]!=64):   # ord(b'@')
            bline = bytes(next(filename),
                          encoding='ascii'
                         )
            
        seqid+=1
        
        cut_space = bline.find(b' ')
        cut_slash = bline.find(b'/')
        if cut_space >= 0 :
            if cut_space < cut_slash or cut_slash < 0:
                sid = bline[1:cut_space]
            else:
                sid = bline[1:cut_slash]
        elif cut_slash >= 0:
            sid = bline[1:cut_slash]
        else:
            sid = bline.strip()
                
        seq = bytes(next(filename),
                    encoding='ascii').strip().upper()
        next(filename)
        quality = array("B",
                        bytes(next(filename),
                              encoding='ascii').strip())
        
        yield(sid,seq,quality)
                



'''
Created on 20 oct. 2012

@author: coissac
'''


from .build_exe import build_exe
from distutils import log

class build_ctools(build_exe):
    description = "build C/C++ executable not distributed with Python extensions"

    def initialize_options(self):
        build_exe.initialize_options(self)
        
        # List of built tools
        self.ctools = None
        self.littlebigman = None


    def finalize_options(self):
        # This might be confusing: both build-cexe and build-temp default
        # to build-temp as defined by the "build" command.  This is because
        # I think that C libraries are really just temporary build
        # by-products, at least from the point of view of building Python
        # extensions -- but I want to keep my options open.
            
        build_exe.finalize_options(self)
        
        
        self.set_undefined_options('build',
                                    ('build_temp', 'build_cexe'))

        self.set_undefined_options('littlebigman',
                                   ('littlebigman',  'littlebigman'))
                   
        self.executables = self.distribution.ctools
        self.check_executable_list(self.executables)
        

        if self.littlebigman =='-DLITTLE_END':
            if self.define is None:
                self.define=[('LITTLE_END',None)]
            else:
                self.define.append('LITTLE_END',None)
                
        log.info('Look for CPU architecture... %s',self.define)

        self.ctools = set()

    def run(self):

        for cmd_name in self.get_sub_commands():
            self.run_command(cmd_name)

        
        build_exe.run(self)
        
        for e,p in self.executables:                                  # @UnusedVariable
            self.ctools.add(e)
            




'''
Created on 2 oct. 2014

@author: coissac
'''

import sys 
import tempfile


from obidistutils.serenity.globals import tmpdir         # @UnusedImport
from obidistutils.serenity.globals import saved_args     # @UnusedImport

def get_serenity_dir():
    global tmpdir

    if not tmpdir:
        tmpdir.append(tempfile.mkdtemp())
    return tmpdir[0]
        
def save_argv():
    global saved_args

    del saved_args[:]
    saved_args.extend(list(sys.argv))

           



'''
Created on 28 sept. 2014

@author: coissac
'''

from orgasm import getOutput
import shutil
import sys
import os

__title__="copy an assembly"


default_config = { "source" : None,
                   "dest"   : None,
                   "force"  : False
                 }

def addOptions(parser):
    parser.add_argument(dest='orgasm:outputfilename',  metavar='outputfilename', 
                        help='name of the assembly to be copied')
    
    parser.add_argument(dest='clone:dest',     metavar='dest', 
                        help='name of the new copy of the assembly' )
    
    parser.add_argument("--force","-f",
                        dest='clone:force',
                        action='store_true',
                        default=None,
                        help='Force cloning even if the destination already exist' )
    


def run(config):
    
    logger=config['orgasm']['logger']
    progress = config['orgasm']['progress']

    source = getOutput(config)
    
    logger.info("Copying the assembly %s to %s" % (config['orgasm']['outputfilename'],
                                                   config['clone']['dest'])) 
    
    if os.path.exists("%s.oas"  % config['clone']['dest']):
        if config['clone']['force']:
            shutil.rmtree("%s.oas"  % config['clone']['dest'])
        else:
            logger.error('Destination assembling exists')
            sys.exit(1)
        
    shutil.copytree("%s.oas"  % config['orgasm']['outputfilename'], 
                    "%s.oas"  % config['clone']['dest'])
    
    



'''
Created on 24 mai 2015

@author: coissac
'''
from distutils.version import StrictVersion
from distutils import sysconfig
import subprocess
import os
import glob
import re

from obidistutils.serenity.checksystem import is_windows_system
import sys


def is_python_version(path=None,minversion='3.4',maxversion=None):
    '''
    Checks that the python version is in the range {minversion,maxversion[
    
        @param path: if None consider the running python
                     otherwise the python pointed by the path
        @param minversion: the minimum version to consider
        @param maxversion: the maximum version to consider (strictly inferior to)
                     
        @return: True if the python version match
        @rtype: bool
    '''
    if path is None:
        pythonversion = StrictVersion(sysconfig.get_python_version())
    else:
        command = """'%s' -c 'from distutils import sysconfig; """ \
                  """print(sysconfig.get_python_version())'""" % path
                  
        p = subprocess.Popen(command, 
                             shell=True, 
                             stdout=subprocess.PIPE)
        pythonversion=str(p.communicate()[0],'utf8').strip()
        pythonversion = StrictVersion(pythonversion)
                        
    return  (        pythonversion >=StrictVersion(minversion) 
             and (   maxversion is None    
                  or pythonversion < StrictVersion(maxversion))
            )
           

def lookfor_good_python(minversion='3.4',maxversion=None,followLink=False):
    '''
    Look for all python interpreters present in the system path that
    match the version constraints.
    
    @param minversion: the minimum version to consider
    @param maxversion: the maximum version to consider (strictly inferior to)
    @param followLink: a boolean value indicating if link must be substituted 
                       by their real path.
                       
    @return: a list of path to interpreters
    '''
    exe = []
    if not is_windows_system():
        paths = os.environ['PATH'].split(os.pathsep)
        for p in paths:
            candidates = glob.glob(os.path.join(p,'python*')) 
            pexe = []
            pythonpat=re.compile('python([0-9]|[0-9]\.[0-9])?$')
            for e in candidates:
                print(e)
                if pythonpat.search(e) is not None:
                    if followLink and os.path.islink(e):
                        e = os.path.realpath(e)
                    if (os.path.isfile(e) and 
                       os.access(e, os.X_OK) and 
                       is_python_version(e,minversion,maxversion)):
                        pexe.append(e)
            exe.extend(set(pexe))
        
    return exe

def is_a_virtualenv_python(path=None):
    '''
    Check if the python is belonging a virtualenv
    
        @param path: the path pointing to the python executable.
                     if path is None then the running python is
                     considered.
        @param path: str or None 
                     
        @return: True if the python belongs a virtualenv
                 False otherwise
        @rtype: bool
                 
    '''
    if path is None:
        rep = sys.base_exec_prefix != sys.exec_prefix
    else:
        command = """'%s' -c 'import sys; print(sys.base_exec_prefix != sys.exec_prefix)'""" % path
        p = subprocess.Popen(command, 
                             shell=True, 
                             stdout=subprocess.PIPE)
        rep = eval(str(p.communicate()[0],'utf8'))
        
    return rep


def which_virtualenv(path=None,full=False):
    '''
    Returns the name of the virtualenv.
        @param path: the path to a python binary or None
                     if you want to consider the running python
        @type path: str or None
                     
        @param full: if set to True, returns the absolute path,
                     otherwise only return a simple directory name
        @type full: bool
                 
        @return: the virtual environment name or None if the
                 path does not belong a virtualenv
        @rtype: str or None
    '''
    if path is None:
        path = sys.executable
    
    if is_a_virtualenv_python(path):
        parts = path.split(os.sep)
        try:
            if full:
                rep = os.sep.join(parts[0:parts.index('bin')])
                rep = os.path.realpath(rep)
            else:
                rep = parts[parts.index('bin')-1]
        except ValueError:
            rep = None
    else:
        rep=None
        
    return rep
        




'''
Created on 20 oct. 2012

@author: coissac
'''

import os

from obidistutils.command.build_exe import build_exe
from distutils import log

import subprocess

class littlebigman(build_exe):
    
    description = "build the littlebigman executable testing endianness of the CPU"


    def initialize_options(self):
        build_exe.initialize_options(self)

        self.littlebigman = None


    def finalize_options(self):
        # This might be confusing: both build-cexe and build-temp default
        # to build-temp as defined by the "build" command.  This is because
        # I think that C libraries are really just temporary build
        # by-products, at least from the point of view of building Python
        # extensions -- but I want to keep my options open.
        
        build_exe.finalize_options(self)
        
        self.set_undefined_options('build',
                                   ('build_temp', 'build_cexe'))

        # self.ctools = self.distribution.ctools
        
        if os.path.isfile("distutils.ext/src/littlebigman.c"):
            self.executables = [('littlebigman',{"sources":["distutils.ext/src/littlebigman.c"]})]
            self.check_executable_list(self.executables)
        else:
            self.executables = []


    def run_littlebigman(self):
        p = subprocess.Popen("'%s'" % os.path.join(self.build_temp,
                                                   'littlebigman'), 
                             shell=True, 
                             stdout=subprocess.PIPE)
        little = p.communicate()[0]
        return little.decode('latin1')

    def run(self):
        build_exe.run(self)
        self.littlebigman=self.run_littlebigman()
        log.info("Your CPU is in mode : %s" % self.littlebigman)
        
        


# cython: language_level=3

from cpython cimport *
from orgasm.fiboheap.fibo cimport *
from libc.stdlib cimport malloc,realloc,free
from cpython.int cimport PyInt_AsSsize_t
from cpython.int cimport PyInt_FromLong
from cpython.int cimport PyInt_FromSsize_t

cdef extern from "inttypes.h":
    ctypedef int int32_t  
    ctypedef int int64_t  

cdef struct IntData:
    FiboNode            node                    
    size_t              priority
    int32_t             data                      

cdef class FibonacciIntHeap:

    cdef FiboTree  _treedat
    cdef dict      _nodes
    cdef IntData** _arenas
    cdef size_t    _arenacount
    cdef IntData*  _last 
    cdef size_t    _chunksize
    cdef size_t    _size

    cdef void _realloc(self)
    cdef void _setNode(self,IntData* node,int32_t data,size_t priority)
    cpdef size_t push(self,int32_t data,size_t priority) except *
    cpdef tuple pop(self)
    cpdef int32_t popInt(self)
    cpdef size_t getPriority(self, int32_t key)   except *
    cpdef size_t setPriority(self, int32_t key,size_t priority)   except *





'''
Created on 27 sept. 2016

@author: coissac
'''
from orgasm import getOutput, getIndex, getSeeds
from orgasm.tango import restoreGraph, estimateFragmentLength, genesincontig,\
    scaffold, pairEndedConnected
    
from orgasm.version import version
import pathlib
import sys

__title__="Restore the intermediate graph saved regularly during the assembling process."

default_config = {}

def addOptions(parser):
    parser.add_argument(dest='orgasm:indexfilename',  metavar='<index>', 
                        help='index root filename (produced by the oa index command)')
    
    parser.add_argument(dest='orgasm:outputfilename',     metavar='<output>', 
                                                          nargs='?', 
                                                          default=None,
                        help='output prefix' )
    
        
    
    parser.add_argument('--back',             dest='orgasm:back', 
                                              metavar='<insert size>',
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='the number of bases taken at the end of '
                             'contigs to jump with pared-ends [default: <estimated>]')



   
def cat(filename):
    with open(filename,'r') as f:
        for line in f:
            print(line)
            
def run(config):

    logger=config['orgasm']['logger']
    output = getOutput(config)


    if not pathlib.Path(output+'.gml').is_file():
        logger.error("No assembly graph available")
        sys.exit(1)
        
    if config['graph']['format']=='gml':
        cat(output+'.gml')
    elif  config['graph']['format']=='gmlpath':
        if pathlib.Path(output+'.path.gml').is_file():
            cat(output+'.path.gml')
        else:
            cat(output+'.gml')
    elif  config['graph']['format']=='lastgraph':
        r = getIndex(config)
        ecoverage,seeds,newprobes = getSeeds(r,config)  # @UnusedVariable
        
        asm = restoreGraph(output+'.oax',r,seeds)
    
        logger.info("Evaluate fragment length")
        
        meanlength,sdlength = estimateFragmentLength(asm)
        
        if meanlength is not None:
            logger.info("Fragment length estimated : %f pb (sd: %f)" % (meanlength,sdlength))
    
        if config['orgasm']['back'] is not None:
            back = config['orgasm']['back']
        elif config['orgasm']['back'] is None and meanlength is not None:
            back = int(meanlength + 4 * sdlength)
            if back > 500:
                back=500
        else:
            back = 300
        compactGraph2lastgraph(asm,logger=logger)




import sys

from distutils import util
from distutils import sysconfig
from distutils import log
from distutils.version import LooseVersion, StrictVersion
import glob
import os
import subprocess
import re 
from distutils.errors import DistutilsError
import tempfile

from importlib.util import spec_from_file_location  # @UnresolvedImport
import zipimport

import argparse

import base64

from .checkpython import is_python_version
                        
                        
from obidistutils.serenity.rerun import  enforce_good_python
from obidistutils.serenity.rerun import rerun_with_anothe_python

from obidistutils.serenity.virtual import serenity_virtualenv
                        
from obidistutils.serenity.checksystem import is_mac_system, \
                                              is_windows_system
                        
from obidistutils.serenity.checkpackage import install_requirements
from obidistutils.serenity.checkpackage import check_requirements

from obidistutils.serenity.util import save_argv
                        
from obidistutils.serenity.snake import snake
                            
    
def serenity_snake(envname,package,version):
    old = log.set_threshold(log.INFO)

    log.info("Installing %s (%s) in serenity mode" % (package,version))

    enforce_good_python()

    virtualpython=serenity_virtualenv(envname,package,version)
    
    if virtualpython!=os.path.realpath(sys.executable):
        log.info("Restarting installation within the %s virtualenv" % (envname))
        rerun_with_anothe_python(virtualpython)
        
    log.info("%s will be installed with python : %s" % (package,virtualpython))
        
    if install_requirements():    
        log.info("Restarting installation with all dependencies ok")
        rerun_with_anothe_python(virtualpython)
    
    log.set_threshold(old)
    
def serenity_assert(version):
    check_requirements()


def is_serenity():
    from obidistutils.serenity.globals import local_serenity
    return local_serenity and local_serenity[0]

def serenity_mode(package,version):
    
    save_argv()

    
    from obidistutils.serenity.globals import saved_args
    from obidistutils.serenity.globals import local_serenity
    

    old = log.set_threshold(log.INFO)
    
    argparser = argparse.ArgumentParser(add_help=False)
    argparser.add_argument('--serenity',
                           dest='serenity', 
                           action='store_true',
                           default=False, 
                           help='Switch the installer in serenity mode. Everythings are installed in a virtualenv')

    argparser.add_argument('--virtualenv',
                           dest='virtual', 
                           type=str,
                           action='store',
                           default="%s-%s" % (package,version), 
                           help='Specify the name of the virtualenv used by the serenity mode [default: %s-%s]' % (package,version))    
    
    args, unknown = argparser.parse_known_args()
    sys.argv = [sys.argv[0]] + unknown
    
    if args.serenity:
        local_serenity.append(True)
        serenity_snake(args.virtual,package,version)
    else:
        local_serenity.append(False)       
    
    log.set_threshold(old)
    
    return args.serenity
    
    
def getVersion(source,main,version):
    path  = os.path.join(source,main,'%s.py' % version)
    spec = spec_from_file_location('version',path)
    return spec.loader.load_module().version.strip()
    



# cython: language_level=3

from ._fibointheap cimport *

cdef int cmpMaxIntFunc(const_FiboNode_ptr_const node1, const_FiboNode_ptr_const node2)

cdef class FibonacciMaxIntHeap(FibonacciIntHeap):


    cpdef tuple max(self)
    cpdef int32_t maxInt(self)





'''
Created on 2 oct. 2014

@author: coissac
'''

import os
import sys
import venv

from distutils.errors import DistutilsError
from .globals import local_virtualenv  # @UnusedImport
from .checkpython import   which_virtualenv,\
                           is_python_version, \
                           is_a_virtualenv_python
                                                

    
    
def serenity_virtualenv(envname,package,version,minversion='3.4',maxversion=None):
        
    #
    # Checks if we are already running under the good virtualenv
    #
    ve = which_virtualenv(full=True)
    if ve == os.path.realpath(envname) and is_python_version(minversion=minversion,maxversion=maxversion):
        return sys.executable
        
    #
    # Check if the virtualenv exist
    # 
        
    python = None
    
    if os.path.isdir(envname):
        python = os.path.join(envname,'bin','python')
        ok = (is_python_version(python,
                                minversion=minversion,
                                maxversion=maxversion) and 
              is_a_virtualenv_python(python))
        
        
        #
        # The virtualenv already exist but it is not ok
        #
        if not ok:
            raise DistutilsError("A virtualenv %s already exists but not with the required python")
                 
    else:
        ok = False
                
        
    #
    # Creates a new virtualenv
    #
    if not ok:
        venv.create(envname, 
                    system_site_packages=False, 
                    clear=True, 
                    symlinks=False, 
                    with_pip=True)
                
        # check the newly created virtualenv
        return serenity_virtualenv(envname,package,version)
    
    return os.path.realpath(python)
    
    
    


'''
Created on 13 fevr. 2014

@author: coissac
'''

from distutils import log
import os


from distutils.errors import DistutilsSetupError

try:
    from Cython.Distutils import build_ext  as ori_build_ext  # @UnresolvedImport
    from Cython.Compiler import Options as cython_options  # @UnresolvedImport
    class build_ext(ori_build_ext):
    
        
        def modifyDocScripts(self):
            build_dir_file=open("doc/sphinx/build_dir.txt","w")
            print(self.build_lib,file=build_dir_file)
            build_dir_file.close()
            
        def initialize_options(self):
            ori_build_ext.initialize_options(self)  # @UndefinedVariable
            self.littlebigman = None
            self.built_files = None
    
        
        def finalize_options(self):
            ori_build_ext.finalize_options(self)  # @UndefinedVariable
    
            self.set_undefined_options('littlebigman',
                                       ('littlebigman',  'littlebigman'))
            
            self.set_undefined_options('build_files',
                                       ('files',  'built_files'))
    
            self.cython_c_in_temp = 1
                       
            if self.littlebigman =='-DLITTLE_END':
                if self.define is None:
                    self.define=[('LITTLE_END',None)]
                else:
                    self.define.append('LITTLE_END',None)
            
        def substitute_sources(self,exe_name,sources):
            """
            Substitutes source file name starting by an @ by the actual
            name of the built file (see --> build_files)
            """
            sources = list(sources)
            for i in range(len(sources)):
                message = "%s :-> %s" % (exe_name,sources[i])
                if sources[i][0]=='@':
                    try:
                        filename = self.built_files[sources[i][1:]]
                    except KeyError:
                        tmpfilename = os.path.join(self.build_temp,sources[i][1:])
                        if os.path.isfile       (tmpfilename):
                            filename = tmpfilename
                        else:
                            raise DistutilsSetupError(
                                 'The %s filename declared in the source '
                                 'files of the program %s have not been '
                                 'built by the installation process' % (sources[i],
                                                                         exe_name))
                    sources[i]=filename
                    log.info("%s changed to %s",message,filename)
                else:
                    log.info("%s ok",message)
    
            return sources
    
        def build_extensions(self):
            # First, sanity-check the 'extensions' list
            
            for ext in self.extensions:
                ext.sources = self.substitute_sources(ext.name,ext.sources)
                
            self.check_extensions_list(self.extensions)
    
            for ext in self.extensions:
                log.info("%s :-> %s",ext.name,ext.sources)
                ext.sources = self.cython_sources(ext.sources, ext)
                self.build_extension(ext)
    
            
        def run(self):
            self.modifyDocScripts()
    
            for cmd_name in self.get_sub_commands():
                self.run_command(cmd_name)
    
            cython_options.annotate = True
            ori_build_ext.run(self)  # @UndefinedVariable
            
    
        def has_files(self):
            return self.distribution.has_files()
    
        def has_executables(self):
            return self.distribution.has_executables()
        
        sub_commands = [('build_files',has_files),
                        ('build_cexe', has_executables)
                        ] + \
                       ori_build_ext.sub_commands

except ImportError:
    from distutils.command import build_ext  # @UnusedImport
           


        


'''
Created on 24 mai 2015

@author: coissac
'''

import sys
import os
from distutils import log
from distutils.errors import DistutilsError


from obidistutils.serenity.globals import saved_args
from obidistutils.serenity.checkpython import is_python_version,\
    lookfor_good_python


def rerun_with_anothe_python(path, minversion='3.4',maxversion=None, fork=False):
        
    if saved_args:
        args = saved_args
    else:
        args = list(sys.argv)
        
          
    assert is_python_version(path,minversion,maxversion), \
           'the selected python is not adapted to the installation of this package'
                   
    args.insert(0, path)
        
    sys.stderr.flush()
    sys.stdout.flush()
    
    if fork:
        log.info('Forking a new install process')
        os.system(' '.join(list(args)))
        log.info('External process ended')
        sys.exit(0)
    else:
        log.info('Install script restarting...')
        os.execv(path,list(args))

def enforce_good_python(minversion='3.4',maxversion=None, fork=False):
    if is_python_version(minversion=minversion,maxversion=maxversion):
        log.info('You are running the good python')
        return True
    
    goodpython = lookfor_good_python(minversion,maxversion)
    
    if not goodpython:
        raise DistutilsError('No good python identified on your system')

    goodpython=goodpython[0]
    
    log.warn("========================================")    
    log.warn("")
    log.warn("    Switching to python : %s" % goodpython)
    log.warn("")
    log.warn("========================================")    

    rerun_with_anothe_python(goodpython)



'''
Created on 28 sept. 2014

@author: coissac
'''

import orgasm.samples

from orgasm import getOutput,getIndex, getSeeds



__title__="Build the set of seed reads"


default_config = { "reformat" : None
                 }

def addOptions(parser):
    parser.add_argument(dest='orgasm:indexfilename',  metavar='index', 
                        help='index root filename (produced by the oa index command)')
    
    parser.add_argument(dest='orgasm:outputfilename',     metavar='output', 
                                                          nargs='?', 
                                                          default=None,
                        help='output prefix' )
    
    
    parser.add_argument('--phiX',              dest='orgasm:phix', 
                                              action='store_true', 
                                              default=None, 
                        help='activate the filtering of Phi-X174 sequences [default]')
    
    parser.add_argument('--phiX-off',              dest='orgasm:phix', 
                                              action='store_false', 
                                              default=None, 
                        help='desactivate the filtering of Phi-X174 sequences')
    
    parser.add_argument('--seeds',            dest ='orgasm:seeds', 
                                              metavar='seeds', 
                                              action='append',
                                              default=[], 
                                              type=str, 
                        help='protein or nucleic seeds; either a fasta file containing '
                        'seed sequences or the name of one of the internal set of seeds '
                        'among %s' % str(list(filter(lambda s: s.startswith('prot') or 
                                                s.startswith('nuc'),dir(orgasm.samples)))))

    parser.add_argument('--kup',              dest='orgasm:kup', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='The word size used to identify the seed reads '
                             '[default: protein=4, DNA=12]')

    parser.add_argument("--reformat",
                        dest="seeds:reformat",
                        action='store_true',
                        default=None,
                        help='Asks for reformatting an old sequence index to the new format'
                       )


def run(config):
    
    logger=config['orgasm']['logger']
    output = getOutput(config)                      # @UnusedVariable

    logger.info("Looking for the seed reads")
    
    r = getIndex(config)
    ecoverage,x,newprobes = getSeeds(r,config)      # @UnusedVariable

    logger.info('Coverage estimated from probe matches at : %d' % ecoverage)
    



'''
Created on 20 oct. 2012

@author: coissac
'''

import os

from distutils.core import Command
from distutils.sysconfig import customize_compiler
from distutils.errors import DistutilsSetupError
from distutils import log
from distutils.ccompiler import show_compilers

class build_exe(Command):
    
    description = "build an executable -- Abstract command "

    user_options = [
        ('build-cexe', 'x',
         "directory to build C/C++ libraries to"),
        ('build-temp', 't',
         "directory to put temporary build by-products"),
        ('debug', 'g',
         "compile with debugging information"),
        ('force', 'f',
         "forcibly build everything (ignore file timestamps)"),
        ('compiler=', 'c',
         "specify the compiler type"),
        ]

    boolean_options = ['debug', 'force']

    help_options = [
        ('help-compiler', None,
         "list available compilers", show_compilers),
        ]

    def initialize_options(self):
        self.build_cexe = None
        self.build_temp = None

        # List of executables to build
        self.executables = None

        # Compilation options for all libraries
        self.include_dirs = None
        self.define = None
        self.undef = None
        self.extra_compile_args = None
        self.debug = None
        self.force = 0
        self.compiler = None
        self.sse = None
        self.built_files=None

    def finalize_options(self):
        # This might be confusing: both build-cexe and build-temp default
        # to build-temp as defined by the "build" command.  This is because
        # I think that C libraries are really just temporary build
        # by-products, at least from the point of view of building Python
        # extensions -- but I want to keep my options open.
        self.set_undefined_options('build',
                                   ('build_temp', 'build_temp'),
                                   ('compiler', 'compiler'),
                                   ('debug', 'debug'),
                                   ('force', 'force'))

        if self.include_dirs is None:
            self.include_dirs = self.distribution.include_dirs or []
            
        if isinstance(self.include_dirs, str):
            self.include_dirs = self.include_dirs.split(os.pathsep)

        self.sse = self.distribution.sse
        
        if self.sse is not None:
            if self.extra_compile_args is None:
                self.extra_compile_args=['-m%s' % self.sse]
            else:
                self.extra_compile_args.append('-m%s' % self.sse)

        # XXX same as for build_ext -- what about 'self.define' and
        # 'self.undef' ?

    def run(self):

        if not self.executables:
            return

        self.mkpath(self.build_cexe)

        # Yech -- this is cut 'n pasted from build_ext.py!
        from distutils.ccompiler import new_compiler
        self.compiler = new_compiler(compiler=self.compiler,
                                     dry_run=self.dry_run,
                                     force=self.force)
        customize_compiler(self.compiler)

        if self.include_dirs is not None:
            self.compiler.set_include_dirs(self.include_dirs)
        if self.define is not None:
            # 'define' option is a list of (name,value) tuples
            for (name,value) in self.define:
                self.compiler.define_macro(name, value)

        if self.undef is not None:
            for macro in self.undef:
                self.compiler.undefine_macro(macro)
                
        self.build_executables(self.executables)


    def check_executable_list(self, executables):
        """Ensure that the list of executables is valid.

        `executable` is presumably provided as a command option 'executables'.
        This method checks that it is a list of 2-tuples, where the tuples
        are (executable_name, build_info_dict).

        Raise DistutilsSetupError if the structure is invalid anywhere;
        just returns otherwise.
        """
        if not isinstance(executables, list):
            raise DistutilsSetupError("'executables' option must be a list of tuples")

        for exe in executables:
            if not isinstance(exe, tuple) and len(exe) != 2:
                raise DistutilsSetupError("each element of 'executables' must a 2-tuple")

            name, build_info = exe

            if not isinstance(name, str):
                raise DistutilsSetupError(
                      "first element of each tuple in 'executables' " 
                      "must be a string (the executables name)")
                
            if '/' in name or (os.sep != '/' and os.sep in name):
                raise DistutilsSetupError(
                       "bad executable name '%s': " 
                       "may not contain directory separators" % exe[0])

            if not isinstance(build_info, dict):
                raise DistutilsSetupError(
                      "second element of each tuple in 'executables' " 
                      "must be a dictionary (build info)")

    def get_executable_names(self):
        # Assume the executables list is valid -- 'check_executable_list()' is
        # called from 'finalize_options()', so it should be!
        if not self.executables:
            return None

        exe_names = []
        for (exe_name, build_info) in self.executables:                     # @UnusedVariable
            exe_names.append(exe_name)
        return exe_names


    def get_source_files(self):
        self.check_executable_list(self.executables)
        filenames = []
        for (exe_name, build_info) in self.executables:                     # @UnusedVariable
            sources = build_info.get('sources')
            if sources is None or not isinstance(sources, (list, tuple)):
                raise DistutilsSetupError(
                       "in 'executables' option (library '%s'), "
                       "'sources' must be present and must be "
                       "a list of source filenames" % exe_name)

            filenames.extend(sources)
        return filenames
    
    def substitute_sources(self,exe_name,sources):
        return list(sources)

    def build_executables(self, executables):
        for (exe_name, build_info) in executables:
            sources = build_info.get('sources')
            if sources is None or not isinstance(sources, (list, tuple)):
                raise DistutilsSetupError(
                       "in 'executables' option (library '%s'), " 
                       "'sources' must be present and must be " 
                       "a list of source filenames" % exe_name)
                
            sources = self.substitute_sources(exe_name,sources)

            log.info("building '%s' program", exe_name)

            # First, compile the source code to object files in the library
            # directory.  (This should probably change to putting object
            # files in a temporary build directory.)
            macros = build_info.get('macros')
            include_dirs = build_info.get('include_dirs')
            extra_args = self.extra_compile_args or []
            
            objects = self.compiler.compile(sources,
                                            output_dir=self.build_temp,
                                            macros=macros,
                                            include_dirs=include_dirs,
                                            extra_postargs=extra_args,
                                            debug=self.debug)

            # Now "link" the object files together into a static library.
            # (On Unix at least, this isn't really linking -- it just
            # builds an archive.  Whatever.)
            self.compiler.link_executable(objects, exe_name,
                                            output_dir=self.build_cexe,
                                            debug=self.debug)

        



#cython: language_level=3, boundscheck=False, wraparound=False, cdivision=True

from cpython.array cimport array
from orgasm.apps.counter cimport ProgressCounter
from orgasm.backtranslate._ahocorasick cimport NucAhoCorasick,AhoCorasick

from docutils.nodes import line

from orgasm.files.universalopener import uopen
from ._fasta import readFasta
from ._fastq import readFastq

from zlib import compress,decompress
from pickle import dumps,loads

import sys
from orgasm.samples.phix import phix

__logger__ = None

cdef dict __stats__ = {}
cdef int  __interval__=50000

def logOrPrint(message,level='info'):
    global __logger__
    print(file=sys.stderr)
    if __logger__ is not None:
        getattr(__logger__, level)(message)
    else:
        print(message,file=sys.stderr)


cpdef tuple cut5prime(tuple sequence, int trimfirst=4):
    """
    Cuts the first *trimfirst* nucleotides of the sequence.
    The returned sequence can be empty
    """
    if len(sequence[1]) > trimfirst:
        return (sequence[0],
                sequence[1][trimfirst:],
                sequence[2][trimfirst:])
    else:
        return (sequence[0],
                b'',
                array("B",[]))
        
        
#cython: initializedcheck=False 
cdef int firstbelow(char[:] qualities, int quality, int shift=33):
    cdef int i =0
    cdef int lq=len(qualities)
    quality+=shift
    
    while(i < lq and qualities[i] >= quality):
        i+=1
        
    if i == lq:
        i=-1
        
    return i

cpdef tuple cut3primeQuality(tuple sequence, int quality, int shift=33):
    cdef int pos = firstbelow(sequence[2],quality,shift)
    
    if pos >= 0:
        return (sequence[0],
                sequence[1][0:pos],
                sequence[2][0:pos])
    else:
        return sequence
           
#cython: initializedcheck=False 
cpdef tuple longestACGT(bytes seq):
    cdef int maxLength = 0
    cdef int maxbegin  = 0
    cdef int maxend    = 0
    cdef int begin     = 0
    cdef int end       = 0
    cdef int length    = 0
    cdef int lseq = len(seq)
    cdef char* cseq = seq
    
    while (end < lseq):
        while (end < lseq and 
               ((cseq[end]==65) or   # A
                (cseq[end]==67) or   # C
                (cseq[end]==71) or   # G
                (cseq[end]==84))):   # T
            end+=1

        length = end - begin

        if length > maxLength:
            maxLength = length 
            maxbegin  = begin
            maxend    = end
        
        end+=1
        begin = end
        
    return (maxLength<lseq,maxbegin,maxend)

#cython: initializedcheck=False 
cpdef int bestWindow(char[:] qualities, int length):
    cdef int begin=-length
    cdef int end  = 0
    cdef int lseq = len(qualities)
    cdef int maxscore=0
    cdef int maxbegin=0
    cdef int cumscore=0
    
    if length >= lseq:
        return 0
        
    while (end < lseq):            
        cumscore+=qualities[end]
        
        if begin >=0:
            cumscore-=qualities[begin]
        
        begin+=1
        end+=1
        
        if begin > 0 and cumscore > maxscore:
            maxscore=cumscore
            maxbegin=begin 
        
    return maxbegin

#cython: initializedcheck=False 
cpdef tuple bestQuality(char[:] qualities, int quality, int shift=33):
    cdef int lseq = len(qualities)  # @DuplicatedSignature
    cdef int begin= 0
    cdef int end  = 0               # @DuplicatedSignature
    cdef int score= 0
    cdef int maxbegin
    cdef int maxend
    cdef int maxscore
     
    shift+=quality
    
    while (end < lseq):
        score+=qualities[end]-shift
        end+=1
        if score <= 0:
            begin = end
            score = 0
        if score >= maxscore:
            maxscore=score
            maxbegin=begin
            maxend  =end
     
    return (lseq > (maxend-maxbegin),maxbegin,maxend)       
    
            
def firstAndNextIterator(line,iterator):
    yield line
    
    for line in iterator:
        yield line
     
            
def PEInterleavedIterator(sequenceIterator):
    cdef tuple forward
    cdef int paircount=0
    
    logOrPrint("Decodes interleaved pair-end reads from a single file")

    for forward in sequenceIterator:
        paircount+=1
        yield (forward,next(sequenceIterator))

    __stats__['paircount']=paircount
        
def PESimulIterator(sequenceIterator):
    cdef tuple forward
    cdef array score 
    cdef int paircount=0
    
    logOrPrint("Simulating pair-end reads from single-end reads")

    
    for forward in sequenceIterator:
        score=array("B",forward[2])
        score.reverse()
        paircount+=1
        yield (forward,
               (forward[0],
                reverseComplement(forward[1]),
                score))

    __stats__['paircount']=paircount
     
def PEIterator(forwardIterator, reverseIterator):
    cdef tuple forward
    cdef int paircount=0

    logOrPrint("Two files pair-end data")

    
    for forward in forwardIterator:
        paircount+=1
        yield (forward,next(reverseIterator))
        
    __stats__['paircount']=paircount
        
def MPIterator(PEiterator):
    cdef tuple forward
    cdef tuple reverse
    
    logOrPrint("Analyze data as mate-pairs")

    for forward,reverse in PEiterator:
        forward[2].reverse()
        reverse[2].reverse()
        yield ((forward[0],
                reverseComplement(forward[1]),
                forward[2]),
               (reverse[0],
                reverseComplement(reverse[1]),
                reverse[2])
              )        
    
        
def PEIdCheckerIterator(PEiterator):
    cdef tuple PE
    cdef int   reject=0
    
    logOrPrint("Checking for badly paired sequence ids")
    
    for PE in PEiterator:
        if PE[0][0]==PE[1][0]:
            yield PE
        else:
            reject+=1
            if reject % __interval__ ==0:
                logOrPrint('%d sequence pairs with non corresponding ids' % reject)
                
    __stats__['bad_ids']=reject
        
   
def PECutter(PEiterator, int trimfirst=4):
    cdef tuple forward
    cdef tuple reverse
    
    logOrPrint("Hard clips the %d first base pairs of each read" % trimfirst)
    
    for forward,reverse in PEiterator:
        yield (cut5prime(forward,
                        trimfirst),
               cut5prime(reverse,
                        trimfirst))
        
def PEACGTCutter(PEiterator, int trimfirst=4):
    cdef tuple forward
    cdef tuple reverse
    cdef int   lf,lr,f,t
    cdef int trimmed=0
    
    logOrPrint("Select the longest region containing only [A,C,G,T]")

    for forward,reverse in PEiterator:
        lf,f,t  = longestACGT(forward[1])
        forward = (forward[0],
                   forward[1][f:t],
                   forward[2][f:t]) 
        
        lr,f,t  = longestACGT(reverse[1])
        reverse = (reverse[0],
                   reverse[1][f:t],
                   reverse[2][f:t]) 
        
        if lf or lr:
            trimmed+=1
            if not trimmed % __interval__:
                logOrPrint("%d sequences ACGT trimmed" % trimmed)
                
        yield (forward,reverse)
        
        __stats__["ACGTtrim"]=trimmed

def PELowQualityTrimmer(PEiterator, int quality, int shift=33): 
    cdef tuple forward
    cdef tuple reverse
    cdef int   lf,lr,f,t
    cdef int   trimmed=0
    
    logOrPrint("Soft clipping bad quality regions (below %d)." % quality)

    
    for forward,reverse in PEiterator:
        lf,f,t   = bestQuality(forward[2],quality,shift)
        if lf:
            forward = (forward[0],
                       forward[1][f:t],
                       forward[2][f:t]) 
        
        lr,f,t   = bestQuality(reverse[2],quality,shift)
        if lr : 
            reverse = (reverse[0],
                       reverse[1][f:t],
                       reverse[2][f:t]) 
            
        if lf or lr:
            trimmed+=1
            if not trimmed % __interval__:
                logOrPrint("%d sequences soft quality trimmed" % trimmed)
                

        yield (forward,reverse)
        __stats__["SoftQualityTrim"]=trimmed


def PEQualityTrimmer(PEiterator, int quality, int shift=33): 
    cdef tuple forward
    cdef tuple reverse
    
    logOrPrint("Hard clipping reads after the first base with a quality below %d." % quality)

    for forward,reverse in PEiterator:
        yield (cut3primeQuality(forward,
                               quality,
                               shift),
               cut3primeQuality(reverse,
                               quality,
                               shift))
        

#cython: initializedcheck=False 
def PELengthEstimate(PEiterator, double threshold=0.9,int minlength=81, bint zip=False):
    cdef tuple pair
    cdef tuple forward
    cdef tuple reverse
    cdef list  store = []
    cdef array histo = array('l',[0] * 1001)
    cdef long[:] chisto=histo
    cdef int   l,lf,lr
    cdef int   fb,rb
    cdef int   readcount
    cdef int   cumsum=0
    cdef int   lpair=0
    cdef int   lcut=1001 
    cdef int   counter=0
    cdef ProgressCounter pc = ProgressCounter(8,
                                              head="Read length estimate",
                                              seconde=0.5,
                                              unit="reads"
                                             )
    
    logOrPrint("Selecting the best length to keep %4.1f%% of the reads." % (threshold * 100))
    logOrPrint("   Minimum length set to %dbp" % minlength)
    
    for pair in  PEiterator:
        counter+=1
        pc(counter)
        lf = len(pair[0][1])
        lr = len(pair[1][1])
        l = lf if lf < lr else lr
        chisto[l]+=1
        
        if zip:
            store.append((l,compress(dumps(pair,protocol=3),9)))
        else:
            store.append((l,pair))
        
    readcount= sum(histo) *  threshold
    
    
    while (lcut >= 0 and cumsum < readcount):
        lcut-=1
        cumsum += chisto[lcut]
        
    print("",file=sys.stderr)
    logOrPrint("Indexing length estimated to : %dbp" % lcut)
    
    if lcut < minlength:
        lcut=minlength
        logOrPrint("Estimated length smaller than : %dbp" % minlength)
        logOrPrint("Indexing length reset to : %dbp" % lcut)

    for lpair,cpair in store:
        if lpair >= lcut:
            if zip:
                forward,reverse = loads(decompress(cpair))
            else:
                forward,reverse = cpair
            fb = bestWindow(forward[2],lcut)
            rb = bestWindow(reverse[2],lcut)

            yield  ((forward[0],
                     forward[1][fb:(fb+lcut)],
                     forward[2][fb:(fb+lcut)]),
                    (reverse[0],
                     reverse[1][rb:(rb+lcut)],
                     reverse[2][rb:(rb+lcut)]))
    

def PELengthCut(PEiterator, int length):
    cdef tuple forward
    cdef tuple reverse
    cdef int   fb,rb

    logOrPrint("Selecting the best subsequence of length %d." % length )

    for forward,reverse in  PEiterator:
        if len(forward[1]) >= length and len(reverse[1]) >= length :
            fb = bestWindow(forward[2],length)
            rb = bestWindow(reverse[2],length)
            
            yield  ((forward[0],
                     forward[1][fb:(fb+length)],
                     forward[2][fb:(fb+length)]),
                    (reverse[0],
                     reverse[1][rb:(rb+length)],
                     reverse[2][rb:(rb+length)]))

        

def PESkipFirstReads(PEI,int skip=0):
    cdef int n = 0
    
    logOrPrint("Skipping the %d first entries..." % skip)
    
    for n in range(skip):
        next(PEI)
        
    logOrPrint("Done.")

    return PEI

def PEMaxReadCount(PEI, int readCount):
    cdef int n 
    
    logOrPrint("Selecting %d good entries..." % readCount)

    for n in range(readCount):
        yield next(PEI)
    
#cython: initializedcheck=False
def PEPhiXCleaner(PEiterator):
    cdef tuple pair
    cdef char* forward
    cdef char* reverse
    cdef list fmatch,rmatch
    cdef int  nf,nr
    cdef int reject=0
    cdef AhoCorasick aho=NucAhoCorasick()
    aho.addSequence(phix["phi-X174"])
    aho.finalize()

    logOrPrint("Sorting out PhiX174 read pairs..." )
    
    for pair in PEiterator:
        forward = pair[0][1]
        reverse = pair[1][1]
        nf=(len(forward)-11) * 2 // 3
        nr=(len(reverse)-11) * 2 // 3
        
        fmatch=aho.match(forward)
        rmatch=aho.match(reverse)
        
        if (len(fmatch[1]) >= nf or len(fmatch[2]) >= nf or
            len(rmatch[1]) >= nr or len(rmatch[2]) >= nr):
            reject+=1
            if reject % __interval__ ==0:
                logOrPrint('%d phiX174 sequence pairs rejected' % reject)
        else:
            yield pair 
    __stats__['phix174']=reject
            

def readSequences(filename):
    
    if isinstance(filename, str):
        filename = uopen(filename)
    
    firstline = next(filename)
    li = firstAndNextIterator(firstline,filename)
    if firstline[0]==">":
        return readFasta(li)
    else:
        return readFastq(li)

def readPairedEnd(filenames,
                  str     mode="PE",
                  bint    checkIds=False,
                  int     cut=4,
                  int     badQualityLimit=28,
                  int     qualityCut=0,
                  int     length=0,
                  double  lengthestimate=0.9,
                  int     minlength=81,
                  int     skip=0,
                  int     maxread=0,
                  bint    phix=True,
                  int     shift=33,
                  bint    zip=False
                  ):
        
    PEI = None
    
    if mode=="MP":
        matepair=True
        mode="PE"
    elif mode=="IMP":
        matepair=True
        mode="IPE"
    else:
        matepair=False
    
    if mode=='PE' and len(filenames) > 1:
        forward = readSequences(filenames[0])
        reverse = readSequences(filenames[1])
        PEI = PEIterator(forward,reverse)
    
    if mode=='SimPE':
        PEI = PESimulIterator(filenames[0])
    
    if mode=='IPE' or (mode=='PE' and len(filenames)==0):
        PEI = PEInterleavedIterator(filenames[0])
        
    if PEI is not None :
        
        if skip > 0:
            PEI = PESkipFirstReads(PEI,skip)
            
        if checkIds:
            PEI = PEIdCheckerIterator(PEI)
        
        if cut > 0:
            PEI = PECutter(PEI,cut)
            
        PEI = PEACGTCutter(PEI)
            
        if badQualityLimit > 0:
            PEI = PELowQualityTrimmer(PEI,badQualityLimit,shift)
            
        if qualityCut > 0:
            PEI = PEQualityTrimmer(PEI,qualityCut,shift)
                    
        if phix:
            PEI = PEPhiXCleaner(PEI)
            
        if length>0 :
            PEI = PELengthCut(PEI,length)
        elif lengthestimate > 0:
            PEI = PELengthEstimate(PEI,lengthestimate,minlength,zip)
            
        if maxread > 0:
            PEI = PEMaxReadCount(PEI,maxread)
            
        if matepair:
            PEI = MPIterator(PEI)
        
        return PEI
    
    raise AttributeError('Cannot find suitable sequence file names')
    
    
cpdef dict getStats():
    return __stats__

cpdef setLogger(logger):
    global __logger__
    __logger__=logger
    


import pickle
import sys

from .indexer import Index

from .backtranslate.fasta import fasta
from . import samples
from orgasm.utils.dna import isDNA  # @UnresolvedImport

import os
import os.path

def getOutput(config):
    '''
    @param config: a configuration object
    @return: the path for the output files
    '''
    logger = config['orgasm']['logger']
    
    dirname = "%s.oas" % config['orgasm']['outputfilename']
    if not os.path.exists(dirname):
        if os.path.exists('%s.oax' % config['orgasm']['outputfilename']):
            
            if config['buildgraph']['reformat']:
                #
                # Reformating outpout according to the new format
                #

                # Try to load the index to test its format
                index=getIndex(config)
                
                os.makedirs(dirname) 
                for extension in ['oax','omx','gml','stats',
                                  'intermediate.gml','.intermediate.oax',
                                  'path.gml',
                                  'path']:
                    if os.path.exists('%s.%s' % (config['orgasm']['outputfilename'],
                                                 extension)):
                        os.renames('%s.%s' % (config['orgasm']['outputfilename'],
                                              extension), 
                                   '%s.oas/assembling.%s' % (config['orgasm']['outputfilename'],
                                                             extension))

                # Try to load the seeds to test its format
                seeds=getSeeds(index, config)
               
                sys.exit(0)
               
            else:
                #
                # Exit with an error because the format is obsolete.
                #
                logger.error("The %s assembly is not stored according to new format" % config['orgasm']['outputfilename'])
                logger.error('Run the oa buildgraph command with the --reformat option')
                sys.exit(1)
        else:
            os.makedirs(dirname)
            
            
        
    return "%s/assembling" % dirname
    

def getIndex(config):
    '''
    
    @param config: a configuration object
    @return: an Indexer instance
    '''
    
    logger=config['orgasm']['logger']    
    output=config['orgasm']['indexfilename']
    dirname="%s.odx" % output
    
    if (   (not os.path.exists(dirname))       and
           (os.path.exists('%s.ofx' % output ) and
            os.path.exists('%s.ogx' % output ) and
            os.path.exists('%s.opx' % output ) and
            os.path.exists('%s.orx' % output )
       )):
        logger.error('Index %s uses the old format please run the command' % output)
        logger.error('    oa index --reformat %s' % output)
        sys.exit(1)
    
    return Index("%s/index" % dirname)


def getProbes(config,noprobe=False):
    '''
    According to the configuration file and the --seeds option this function return the set
    of sequence probes to use
    
     
    @param config: a configuration object
    @return: a dictionary with gene name as key and the 
             nucleic or protein sequence as value
    @rtype: dict
    '''
    logger=config['orgasm']['logger']
    
    if noprobe:
        seeds =config['orgasm']['noseeds']
    else:
        seeds =config['orgasm']['seeds']
        
    
    probes = {}
    
    if seeds is not None:
    
        for s in seeds:
            try:
                p = fasta(s)
                logger.info("Load probe sequences from file : %s" % s)
            except IOError:
                p = getattr(samples,s)
                logger.info("Load probe internal dataset : %s" % s)
                
            probes[s]=[p,{}]
            
    else:
        logger.info("No new probe set specified")
                            
    return probes

def getAdapters(config):
    '''
    According to the configuration file and the --seeds option this function return the set
    of sequence probes to use
    
     
    @param config: a configuration object
    @return: a dictionary with gene name as key and the 
             nucleic or protein sequence as value
    @rtype: dict
    '''
    logger=config['orgasm']['logger']
    adapt5 =config['orgasm']['adapt5']
    adapt3 =config['orgasm']['adapt3']
    
    try:
        p3 = fasta(adapt3).values()
        logger.info("Load 3' adapter sequences from file : %s" % adapt3)
    except IOError:
        p3 = getattr(samples,adapt3)
        logger.info("Load 3' adapter internal dataset : %s" % adapt3)
                    
    try:
        p5 = fasta(adapt5).values()
        logger.info("Load 5' adapter sequences from file : %s" % adapt5)
    except IOError:
        p5 = getattr(samples,adapt5)
        logger.info("Load 5' adapter internal dataset : %s" % adapt5)
                    
    return (p3,p5)


def getSeeds(index,config):
    # looks if the internal blast was already run            
    output = config['orgasm']['outputfilename'] 
    logger=config['orgasm']['logger']
    kup=-1 if config['orgasm']['kup'] is None else config['orgasm']['kup']
    clean=config['buildgraph']['clean']
    forceseeds=config['buildgraph']['forceseeds']
    
    filename="%s.oas/assembling.omx" % output
    
    #
    # Check if the seeds are not correctly placed in the oas directory
    #

    oldfilename=output+'.omx'
    
    if (not os.path.exists(filename) and 
        os.path.exists(oldfilename)):
        if config['seeds']['reformat']:
            os.renames(oldfilename, filename)
        else:
            logger.error('Seed matches are not stored following the new format')
            logger.error('Run the oa seeds command with the --reformat option')
            sys.exit(1)
    
    
    #
    # Look for seeds with the new probe sets
    #

    newprobes  = getProbes(config,noprobe=False)
    newnoprobes= getProbes(config,noprobe=True)
    
    # 
    # Load already run probe sets
    #
    reformated=[]
    if not clean or not newprobes:
        try:
            with open(filename,'rb') as fseeds:
                probes = pickle.load(fseeds)
            logger.info("Load matches from previous run : %d probe sets restored" % len(probes))
    
            if not isinstance(list(probes.values())[0][0], dict):
                logger.error("Too old version of probes matches that cannot be reformated")
                logger.error("Generate a new probe match set using the command oa seeds --seeds command")
                logger.error("The old unsuable seed match has been erased")
                os.remove(filename)
                sys.exit(1)
                
            oldversion=False
            for probename in probes:
                s = probes[probename][1]
                if len(s[list(s.keys())[0]][0]) != 6:
                    if config['seeds']['reformat']:
                        oldversion=True
        
                        logger.warning("Old probe version save on the disk. Recomputes probes %s" % probename)
                        
                        p = probes[probename][0]
                        logger.info("    -> probe set: %s" % probename)
             
                        seeds = index.lookForSeeds(p,
                                                   mincov=config['orgasm']['seedmincov'],
                                                   kup=kup,
                                                   identity=config['orgasm']['identity'],
                                                   logger=logger)
                        
                        probes[probename][1]=seeds
                        
                        reformated.append(probename)
                    
                        logger.info("==> %d matches" % sum(len(seeds[i]) for i in seeds))
                    else:
                        logger.error('Seed matches are not stored following the new format')
                        logger.error('Run the oa seeds command with the --reformat option')
                        sys.exit(1)
            
            if oldversion:
                with open(filename,"wb") as fseeds:
                    pickle.dump(probes,fseeds)
    
            nm=0
            for k in probes:
                for m in probes[k][1].values():
                    nm+=len(m)
            logger.info("   ==> A total of : %d" % nm)
        except FileNotFoundError: 
            logger.info("No previous matches loaded")
            probes={}
    else:
        if os.path.exists(filename):
            logger.info("Cleaning previous matches")
            os.remove(filename)
        probes={}

    
    if newprobes:
        # --seeds option on the command line -> look for these seeds
        logger.info("Running probes matching against reads...")
        
        for probename in newprobes:
            p = newprobes[probename][0]
            logger.info("    -> probe set: %s" % probename)
 
            seeds = index.lookForSeeds(p,
                                        mincov=config['orgasm']['seedmincov'],
                                        kup=kup,
                                        identity=config['orgasm']['identity'],
                                        logger=logger)
            
            nmatches = sum(len(seeds[i]) for i in seeds)
                        
            logger.info("==> %d matches" % nmatches)
            
            if nmatches:            
                probes[probename]=[p,seeds]
                
    noprobes = set()
    if newnoprobes:
        # --seeds option on the command line -> look for these seeds
        logger.info("Running no-probes matching against reads...")
        
        for probename in newnoprobes:
            p = newnoprobes[probename][0]
            logger.info("    -> probe set: %s" % probename)
 
            seeds = index.lookForSeeds(p,
                                        mincov=config['orgasm']['seedmincov'],
                                        kup=kup,
                                        identity=config['orgasm']['identity'],
                                        logger=logger)
            
            nmatches = sum(len(seeds[i]) for i in seeds)
                        
            logger.info("==> %d matches" % nmatches)
            
            if nmatches:   
                for vs in seeds.values():
                    noprobes|=set([x[0] for x in vs])
        
    
    if noprobes:
        logger.info("Removing no-probes matches...")
        for s in probes:
            print(s)
            for p in probes[s][1]:
                probes[s][1][p]=[x for x in probes[s][1][p] if x[0] not in noprobes]
                if not probes[s][1][p]:
                    del probes[s][1][p]
            if not probes[s][1]:
                del probes[s]
        logger.info("Done.")  
    
    newprobes = list(newprobes.keys()) 
    newnoprobes = list(newnoprobes.keys()) 
     
    if not probes:
        logger.info("No --seeds option specified and not previous matches stored")
        sys.exit(1)
        
        
    if newprobes or newnoprobes:
        with open(filename,"wb") as fseeds:
            pickle.dump(probes,fseeds)

    logger.info("Match list :") 

    covmax=0

    for probename in probes:
        p = probes[probename][0]
        s = probes[probename][1]
        nuc = all([isDNA(k) for k in p.values()])
        #print(s[list(s.keys())[0]])
        nbmatch = [(k,
                    sum(x[2] for x in s[k]),
                    sum(x[2] for x in s[k])*index.getReadSize() / len(p[k]) / (1 if nuc else 3)) for k in s]
                 
        nbmatch.sort(key=lambda x:-x[2])
        
        for gene, nb, cov in nbmatch:
            logger.info("     %-10s : %5d (%5.1fx)" % (gene,nb,cov))
        coverage=nbmatch[0][2]
        if coverage > covmax:
            covmax=coverage
        
    if forceseeds:
        newprobes=None

    return covmax,probes,newprobes

#def reloadAssembling